# -*- coding: utf-8 -*-
import re
import os
import sys

import traceback

from DisplayCAL.config import (
    defaults,
    fs_enc,
    get_argyll_display_number,
    get_data_path,
    get_display_profile,
    get_display_rects,
    getbitmap,
    getcfg,
    geticon,
    get_verified_path,
    profile_ext,
    setcfg,
    writecfg,
)
from DisplayCAL.meta import name as appname
from DisplayCAL.options import debug
from DisplayCAL.util_dict import dict_slice, dict_sort
from DisplayCAL.util_io import GzipFileProper
from DisplayCAL.util_list import intlist
from DisplayCAL.util_os import launch_file, make_win32_compatible_long_path, waccess
from DisplayCAL.util_str import strtr, universal_newlines, wrap
from DisplayCAL.worker import (
    Error,
    UnloggedError,
    check_set_argyll_bin,
    get_argyll_util,
    make_argyll_compatible_path,
    show_result_dialog,
)
from DisplayCAL.wxaddons import get_platform_window_decoration_size, wx
from DisplayCAL.wxLUTViewer import LUTCanvas, LUTFrame
from DisplayCAL.wxVRML2X3D import vrmlfile2x3dfile
from DisplayCAL.wxwindows import (
    BaseApp,
    BaseFrame,
    BitmapBackgroundPanelText,
    BitmapBackgroundPanelTextGamut,
    CustomCheckBox,
    CustomGrid,
    CustomRowLabelRenderer,
    FileDrop,
    SimpleBook,
    TwoWaySplitter,
)
from DisplayCAL.wxfixes import GenBitmapButton as BitmapButton, wx_Panel, set_maxsize
from DisplayCAL import colormath
from DisplayCAL import config
from DisplayCAL import wxenhancedplot as plot
from DisplayCAL import localization as lang
from DisplayCAL import ICCProfile as ICCP
from DisplayCAL import x3dom

BGCOLOUR = "#333333"
FGCOLOUR = "#999999"
TEXTCOLOUR = "#333333"

if sys.platform == "darwin":
    FONTSIZE_SMALL = 10
else:
    FONTSIZE_SMALL = 8


class GamutCanvas(LUTCanvas):
    def __init__(self, *args, **kwargs):
        LUTCanvas.__init__(self, *args, **kwargs)
        self.SetEnableTitle(False)
        self.SetFontSizeAxis(FONTSIZE_SMALL)
        self.SetFontSizeLegend(FONTSIZE_SMALL)
        self.pcs_data = []
        self.profiles = {}
        self.colorspace = "a*b*"
        self.intent = ""
        self.direction = ""
        self.order = ""
        self.reset()
        self.resetzoom()

    def DrawCanvas(
        self,
        title=None,
        colorspace=None,
        whitepoint=None,
        center=False,
        show_outline=True,
    ):
        if not title:
            title = ""
        if colorspace:
            self.colorspace = colorspace

        # Defaults
        poly = plot.PolyLine
        poly._attributes["width"] = 3
        polys = []

        # L*a*b*
        optimalcolors = colormath.optimalcolors_Lab

        # CIE 1931 2-deg chromaticity coordinates
        # http://www.cvrl.org/offercsvccs.php
        xy = colormath.cie1931_2_xy

        if self.colorspace == "xy":
            label_x = "x"
            label_y = "y"
            if show_outline:
                polys.append(
                    plot.PolySpline(
                        xy, colour=wx.Colour(102, 102, 102, 153), width=1.75
                    )
                )
                polys.append(
                    plot.PolyLine(
                        [xy[0], xy[-1]],
                        colour=wx.Colour(102, 102, 102, 153),
                        width=1.75,
                    )
                )
            max_x = 0.75
            max_y = 0.85
            min_x = -0.05
            min_y = -0.05
            step = 0.1
        elif self.colorspace == "u'v'":
            label_x = "u'"
            label_y = "v'"
            if show_outline:
                uv = [colormath.xyY2Lu_v_(x, y, 100)[1:] for x, y in xy]
                polys.append(
                    plot.PolySpline(
                        uv, colour=wx.Colour(102, 102, 102, 153), width=1.75
                    )
                )
                polys.append(
                    plot.PolyLine(
                        [uv[0], uv[-1]],
                        colour=wx.Colour(102, 102, 102, 153),
                        width=1.75,
                    )
                )
            max_x = 0.625
            max_y = 0.6
            min_x = -0.025
            min_y = -0.025
            step = 0.1
        elif self.colorspace == "u*v*":
            # Hard to present gamut projection appropriately in 2D
            # because blue tones 'cave in' towards the center
            label_x = "u*"
            label_y = "v*"
            max_x = 150.0
            max_y = 150.0
            min_x = -150.0
            min_y = -150.0
            step = 50
        elif self.colorspace == "DIN99":
            label_x = "a99"
            label_y = "b99"
            max_x = 50.0
            max_y = 50.0
            min_x = -50.0
            min_y = -50.0
            step = 25
        elif self.colorspace in ("DIN99b", "DIN99c", "DIN99d"):
            label_x = f"a{self.colorspace[3:]}"
            label_y = f"b{self.colorspace[3:]}"
            max_x = 65.0
            max_y = 65.0
            min_x = -65.0
            min_y = -65.0
            step = 25
        elif self.colorspace == "ICtCp":
            label_x = "Ct"
            label_y = "Cp"
            max_x = 0.4
            max_y = 0.5
            min_x = -0.5
            min_y = -0.4
            step = 0.1
        elif self.colorspace == "IPT":
            label_x = "P"
            label_y = "T"
            max_x = 1.0
            max_y = 1.0
            min_x = -1.0
            min_y = -1.0
            step = 0.25
        else:
            if self.colorspace == "Lpt":
                label_x = "p"
                label_y = "t"
            else:
                label_x = "a*"
                label_y = "b*"
            max_x = 150.0
            max_y = 150.0
            min_x = -150.0
            min_y = -150.0
            step = 50

        convert2coords = {
            "a*b*": lambda X, Y, Z: colormath.XYZ2Lab(*[v * 100 for v in (X, Y, Z)])[
                1:
            ],
            "xy": lambda X, Y, Z: colormath.XYZ2xyY(X, Y, Z)[:2],
            "u*v*": lambda X, Y, Z: colormath.XYZ2Luv(*[v * 100 for v in (X, Y, Z)])[
                1:
            ],
            "u'v'": lambda X, Y, Z: colormath.XYZ2Lu_v_(X, Y, Z)[1:],
            "DIN99": lambda X, Y, Z: colormath.XYZ2DIN99(*[v * 100 for v in (X, Y, Z)])[
                1:
            ],
            "DIN99b": lambda X, Y, Z: colormath.XYZ2DIN99b(
                *[v * 100 for v in (X, Y, Z)]
            )[1:],
            "DIN99c": lambda X, Y, Z: colormath.XYZ2DIN99c(
                *[v * 100 for v in (X, Y, Z)]
            )[1:],
            "DIN99d": lambda X, Y, Z: colormath.XYZ2DIN99d(
                *[v * 100 for v in (X, Y, Z)]
            )[1:],
            "ICtCp": lambda X, Y, Z: colormath.XYZ2ICtCp(X, Y, Z, clamp=False)[1:],
            "IPT": lambda X, Y, Z: colormath.XYZ2IPT(X, Y, Z)[1:],
            "Lpt": lambda X, Y, Z: colormath.XYZ2Lpt(*[v * 100 for v in (X, Y, Z)])[1:],
        }[self.colorspace]

        if show_outline and (
            self.colorspace == "a*b*" or self.colorspace.startswith("DIN99")
        ):
            if self.colorspace == "a*b*":
                optimalcolors = [Lab[1:] for Lab in optimalcolors]
            else:
                optimalcolors = [
                    convert2coords(*colormath.Lab2XYZ(L, a, b))
                    for L, a, b in optimalcolors
                ]
            polys.append(
                plot.PolySpline(
                    optimalcolors, colour=wx.Colour(102, 102, 102, 153), width=1.75
                )
            )

        # Add color temp graph
        if whitepoint == 1:
            colortemps = []
            for kelvin in range(4000, 25001, 40):
                colortemps.append(convert2coords(*colormath.CIEDCCT2XYZ(kelvin)))
            polys.append(
                plot.PolyLine(
                    colortemps, colour=wx.Colour(255, 255, 255, 204), width=1.5
                )
            )
        elif whitepoint == 2:
            colortemps = []
            # Stepsize 38 = endpoint 24999
            for kelvin in range(1667, 25001, 38):
                colortemps.append(convert2coords(*colormath.planckianCT2XYZ(kelvin)))
            polys.append(
                plot.PolyLine(
                    colortemps, colour=wx.Colour(255, 255, 255, 204), width=1.5
                )
            )

        kwargs = {"scale": 255}

        amount = len(self.pcs_data)

        for i, pcs_triplets in enumerate(reversed(self.pcs_data)):
            if not pcs_triplets or len(pcs_triplets) == 1:
                amount -= 1
                continue

            # Convert xicclu output to coordinates
            coords = []
            for pcs_triplet in pcs_triplets:
                coords.append(convert2coords(*pcs_triplet))

            profile = self.profiles.get(len(self.pcs_data) - 1 - i)
            if (
                profile
                and profile.profileClass == b"nmcl"
                and "ncl2" in profile.tags
                and isinstance(profile.tags.ncl2, ICCP.NamedColor2Type)
                and profile.connectionColorSpace in (b"Lab", b"XYZ")
            ):
                # Named color profile
                for j, (x, y) in enumerate(coords):
                    RGBA = colormath.XYZ2RGB(*pcs_triplets[j], **kwargs)
                    polys.append(
                        plot.PolyMarker(
                            [(x, y)],
                            colour=wx.Colour(*intlist(RGBA)),
                            size=2,
                            marker="plus",
                            width=1.75,
                        )
                    )
            else:
                xy = []
                for x, y in coords[:-1]:
                    xy.append((x, y))
                    if i == 0 or amount == 1:
                        if x > max_x:
                            max_x = x
                        if y > max_y:
                            max_y = y
                        if x < min_x:
                            min_x = x
                        if y < min_y:
                            min_y = y

                xy2 = []
                for j, (x, y) in enumerate(xy):
                    xy2.append((x, y))
                    if len(xy2) == self.size:
                        xy3 = []
                        for k, (x, y) in enumerate(xy2):
                            xy3.append((x, y))
                            if len(xy3) == 2:
                                if i == 1:
                                    # Draw comparison profile with grey outline
                                    RGBA = 102, 102, 102, 255
                                    w = 2
                                else:
                                    RGBA = colormath.XYZ2RGB(
                                        *pcs_triplets[j - len(xy2) + k], **kwargs
                                    )
                                    w = 3
                                polys.append(
                                    poly(
                                        list(xy3),
                                        colour=wx.Colour(*intlist(RGBA)),
                                        width=w,
                                    )
                                )
                                if i == 1:
                                    xy3 = []
                                else:
                                    xy3 = xy3[1:]
                        xy2 = xy2[self.size:]

            # Add whitepoint
            x, y = coords[-1]
            if i == 1:
                # Draw comparison profile with grey outline
                RGBA = 204, 204, 204, 102
                marker = "cross"
                s = 1.5
                w = 1.75
            else:
                RGBA = colormath.XYZ2RGB(*pcs_triplets[-1], **kwargs)
                marker = "plus"
                s = 2
                w = 1.75
            polys.append(
                plot.PolyMarker(
                    [(x, y)],
                    colour=wx.Colour(*intlist(RGBA)),
                    size=s,
                    marker=marker,
                    width=w,
                )
            )

        xy_range = max(abs(min_x), abs(min_y)) + max(max_x, max_y)
        if center or not self.last_draw:
            self.axis_x = self.axis_y = (min(min_x, min_y), max(max_x, max_y))
            self.spec_x = xy_range / step
            self.spec_y = xy_range / step

        if polys:
            graphics = plot.PlotGraphics(polys, title, label_x, label_y)
            p1, p2 = graphics.boundingBox()
            spacer = plot.PolyMarker(
                [
                    (p1[0] - xy_range / 20.0, p1[1] - xy_range / 20.0),
                    (p2[0] + xy_range / 20.0, p2[1] + xy_range / 20.0),
                ],
                colour=wx.Colour(0x33, 0x33, 0x33),
                size=0,
            )
            graphics.objects.append(spacer)
            if center or not self.last_draw:
                boundingbox = spacer.boundingBox()
                self.resetzoom(boundingbox)
                self.center(boundingbox)
            self._DrawCanvas(graphics)

    def reset(self):
        self.axis_x = self.axis_y = -128, 128

    def set_pcs_data(self, i):
        if len(self.pcs_data) < i + 1:
            self.pcs_data.append([])
        else:
            self.pcs_data[i] = []

    def setup(
        self, profiles=None, profile_no=None, intent="a", direction="f", order="n"
    ):
        self.size = (
            40  # Number of segments from one primary to the next secondary color
        )

        if not check_set_argyll_bin():
            return

        # Setup xicclu
        xicclu = get_argyll_util("xicclu")
        if not xicclu:
            return
        cwd = self.worker.create_tempdir()
        if isinstance(cwd, Exception):
            raise cwd

        if not profiles:
            profiles = [
                ICCP.ICCProfile(get_data_path("ref/sRGB.icm")),
                get_display_profile(),
            ]
        for i, profile in enumerate(profiles):
            if profile_no is not None and i != profile_no:
                continue

            if (
                not profile
                or profile.profileClass == b"link"
                or profile.connectionColorSpace not in (b"Lab", b"XYZ")
            ):
                self.set_pcs_data(i)
                self.profiles[i] = None
                continue

            check = self.profiles.get(i)
            if (
                check
                and check.fileName == profile.fileName
                and check.ID == profile.calculateID(False)
                and intent == self.intent
                and direction == self.direction
                and order == self.order
            ):
                continue

            self.profiles[i] = profile

            self.set_pcs_data(i)

            pcs_triplets = []
            if (
                profile.profileClass == b"nmcl"
                and "ncl2" in profile.tags
                and isinstance(profile.tags.ncl2, ICCP.NamedColor2Type)
                and profile.connectionColorSpace in (b"Lab", b"XYZ")
            ):
                for k in profile.tags.ncl2:
                    v = profile.tags.ncl2[k]
                    color = list(v.pcs.values())
                    if profile.connectionColorSpace == b"Lab":
                        # Need to convert to XYZ
                        color = colormath.Lab2XYZ(*color)
                    if intent == "a" and "wtpt" in profile.tags:
                        color = colormath.adapt(
                            color[0],
                            color[1],
                            color[2],
                            whitepoint_destination=list(profile.tags.wtpt.ir.values()),
                        )
                    pcs_triplets.append(color)
                pcs_triplets.sort()
            elif profile.version >= 4 and not profile.convert_iccv4_tags_to_iccv2():
                self.profiles[i] = None
                self.errors.append(
                    Error(
                        "\n".join(
                            [
                                lang.getstr("profile.iccv4.unsupported"),
                                profile.getDescription(),
                            ]
                        )
                    )
                )
                continue
            else:
                channels = {
                    b"XYZ": 3,
                    b"Lab": 3,
                    b"Luv": 3,
                    b"YCbr": 3,
                    b"Yxy": 3,
                    b"RGB": 3,
                    b"GRAY": 1,
                    b"HSV": 3,
                    b"HLS": 3,
                    b"CMYK": 4,
                    b"CMY": 3,
                    b"2CLR": 2,
                    b"3CLR": 3,
                    b"4CLR": 4,
                    b"5CLR": 5,
                    b"6CLR": 6,
                    b"7CLR": 7,
                    b"8CLR": 8,
                    b"9CLR": 9,
                    b"ACLR": 10,
                    b"BCLR": 11,
                    b"CCLR": 12,
                    b"DCLR": 13,
                    b"ECLR": 14,
                    b"FCLR": 15,
                }.get(profile.colorSpace)

                if not channels:
                    self.errors.append(
                        Error(
                            lang.getstr(
                                "profile.unsupported",
                                (profile.profileClass, profile.colorSpace),
                            )
                        )
                    )
                    continue

                # Create input values
                device_values = []
                if profile.colorSpace in (b"Lab", b"Luv", b"XYZ", b"Yxy"):
                    # Use ICC PCSXYZ encoding range
                    minv = 0.0
                    maxv = 0xFFFF / 32768.0
                else:
                    minv = 0.0
                    maxv = 1.0
                step = (maxv - minv) / (self.size - 1)
                for j in range(min(3, channels)):
                    for k in range(min(3, channels)):
                        device_value = [0.0] * channels
                        device_value[j] = maxv
                        if j != k or channels == 1:
                            for l in range(self.size):
                                device_value[k] = minv + step * l
                                device_values.append(list(device_value))
                if profile.colorSpace in (
                    b"HLS",
                    b"HSV",
                    b"Lab",
                    b"Luv",
                    b"YCbr",
                    b"Yxy",
                ):
                    # Convert to actual color space
                    # TODO: Handle HLS and YCbr
                    tmp = list(device_values)
                    device_values = []
                    for j, values in enumerate(tmp):
                        if profile.colorSpace == b"HSV":
                            HSV = list(colormath.RGB2HSV(*values))
                            device_values.append(HSV)
                        elif profile.colorSpace == b"Lab":
                            Lab = list(colormath.XYZ2Lab(*[v * 100 for v in values]))
                            device_values.append(Lab)
                        elif profile.colorSpace == b"Luv":
                            Luv = list(colormath.XYZ2Luv(*[v * 100 for v in values]))
                            device_values.append(Luv)
                        elif profile.colorSpace == b"Yxy":
                            xyY = list(colormath.XYZ2xyY(*values))
                            device_values.append(xyY)

                # Add white
                if profile.colorSpace == b"RGB":
                    device_values.append([1.0] * channels)
                elif profile.colorSpace == b"HLS":
                    device_values.append([0, 1, 0])
                elif profile.colorSpace == b"HSV":
                    device_values.append([0, 0, 1])
                elif profile.colorSpace in (b"Lab", b"Luv", b"YCbr"):
                    if profile.colorSpace == b"YCbr":
                        device_values.append([1.0, 0.0, 0.0])
                    else:
                        device_values.append([100.0, 0.0, 0.0])
                elif profile.colorSpace in (b"XYZ", b"Yxy"):
                    if profile.colorSpace == b"XYZ":
                        device_values.append(list(profile.tags.wtpt.pcs.values()))
                    else:
                        device_values.append(profile.tags.wtpt.pcs.xyY)
                elif profile.colorSpace != b"GRAY":
                    device_values.append([0.0] * channels)

                if debug:
                    print("In:")
                    for v in device_values:
                        print(" ".join(("%3.4f",) * len(v)) % tuple(v))

                # Lookup device -> XYZ values through profile using xicclu
                if direction == "ib" and intent not in "ar":
                    fwd_intent = "r"
                else:
                    fwd_intent = intent
                try:
                    # Device -> PCS, fwd
                    odata = self.worker.xicclu(
                        profile, device_values, intent, "f", order
                    )
                    if direction == "ib":
                        # PCS -> device, bwd
                        odata = self.worker.xicclu(profile, odata, intent, "b", order)
                        # Device -> PCS, fwd
                        odata = self.worker.xicclu(
                            profile, odata, fwd_intent, "f", order
                        )
                except Exception as exception:
                    self.errors.append(Error(str(exception)))
                    continue

                if debug:
                    print("Out:")
                for pcs_triplet in odata:
                    if debug:
                        print(
                            " ".join(("%3.4f",) * len(pcs_triplet)) % tuple(pcs_triplet)
                        )
                    pcs_triplets.append(pcs_triplet)
                    if profile.connectionColorSpace == b"Lab":
                        pcs_triplets[-1] = list(colormath.Lab2XYZ(*pcs_triplets[-1]))

            if len(self.pcs_data) < i + 1:
                self.pcs_data.append(pcs_triplets)
            else:
                self.pcs_data[i] = pcs_triplets

        # Remove temporary files
        self.worker.wrapup(False)

        self.intent = intent
        self.direction = direction
        self.order = order


class GamutViewOptions(wx_Panel):
    def __init__(self, *args, **kwargs):
        scale = getcfg("app.dpi") / config.get_default_dpi()
        if scale < 1:
            scale = 1

        super(GamutViewOptions, self).__init__(*args, **kwargs)
        self.SetBackgroundColour(BGCOLOUR)
        self.sizer = wx.FlexGridSizer(0, 3, 4, 0)
        self.sizer.AddGrowableCol(0)
        self.sizer.AddGrowableCol(1)
        self.sizer.AddGrowableCol(2)
        self.SetSizer(self.sizer)

        self.sizer.Add((0, 0))
        legendsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(legendsizer)

        # Whitepoint legend
        legendsizer.Add(
            wx.StaticBitmap(self, -1, getbitmap("theme/cross-2px-12x12-fff")),
            flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT,
            border=2,
        )
        self.whitepoint_legend = wx.StaticText(self, -1, lang.getstr("whitepoint"))
        self.whitepoint_legend.SetMaxFontSize(11)
        self.whitepoint_legend.SetForegroundColour(FGCOLOUR)
        legendsizer.Add(
            self.whitepoint_legend, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=10
        )

        # Comparison profile whitepoint legend
        self.comparison_whitepoint_bmp = wx.StaticBitmap(
            self, -1, getbitmap("theme/x-2px-12x12-999")
        )
        legendsizer.Add(
            self.comparison_whitepoint_bmp,
            flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT,
            border=20,
        )
        self.comparison_whitepoint_legend = wx.StaticText(
            self,
            -1,
            "{} ({})".format(
                lang.getstr("whitepoint"), lang.getstr("comparison_profile")
            ),
        )
        self.comparison_whitepoint_legend.SetMaxFontSize(11)
        self.comparison_whitepoint_legend.SetForegroundColour(FGCOLOUR)
        legendsizer.Add(
            self.comparison_whitepoint_legend,
            flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT,
            border=8,
        )
        self.sizer.Add((0, 0))

        # Empty 'row'
        self.sizer.Add((0, 0))
        self.sizer.Add((0, 0))
        self.sizer.Add((0, 0))

        self.sizer.Add((0, 0))
        self.options_sizer = wx.FlexGridSizer(0, 3, 4, 8)
        self.options_sizer.AddGrowableCol(2)
        self.sizer.Add(self.options_sizer, flag=wx.EXPAND)

        # Colorspace select
        self.colorspace_outline_bmp = wx.StaticBitmap(
            self, -1, getbitmap("theme/solid-16x2-666")
        )
        self.options_sizer.Add(
            self.colorspace_outline_bmp, flag=wx.ALIGN_CENTER_VERTICAL
        )
        self.colorspace_label = wx.StaticText(self, -1, lang.getstr("colorspace"))
        self.colorspace_label.SetMaxFontSize(11)
        self.colorspace_label.SetForegroundColour(FGCOLOUR)
        self.options_sizer.Add(self.colorspace_label, flag=wx.ALIGN_CENTER_VERTICAL)
        self.colorspace_select = wx.Choice(
            self,
            -1,
            size=(150 * scale, -1),
            choices=[
                "CIE a*b*",
                "CIE u*v*",
                "CIE u'v'",
                "CIE xy",
                "DIN99",
                "DIN99b",
                "DIN99c",
                "DIN99d",
                "ICtCp",
                "IPT",
                "Lpt",
            ],
        )
        self.options_sizer.Add(
            self.colorspace_select, flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND
        )
        self.colorspace_select.Bind(wx.EVT_CHOICE, self.generic_select_handler)
        self.colorspace_select.SetSelection(0)

        # Colorspace outline
        self.options_sizer.Add((0, 0))
        self.options_sizer.Add((0, 0))
        self.draw_gamut_outline_cb = CustomCheckBox(
            self, -1, lang.getstr("colorspace.show_outline")
        )
        self.draw_gamut_outline_cb.Bind(
            wx.EVT_CHECKBOX, self.draw_gamut_outline_handler
        )
        self.draw_gamut_outline_cb.SetMaxFontSize(11)
        self.draw_gamut_outline_cb.SetForegroundColour(FGCOLOUR)
        self.draw_gamut_outline_cb.SetValue(True)
        self.options_sizer.Add(
            self.draw_gamut_outline_cb,
            flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT,
            border=4,
        )

        # Colortemperature curve select
        self.whitepoint_bmp = wx.StaticBitmap(
            self, -1, getbitmap("theme/solid-16x1-fff")
        )
        self.options_sizer.Add(self.whitepoint_bmp, flag=wx.ALIGN_CENTER_VERTICAL)
        self.whitepoint_label = wx.StaticText(
            self, -1, lang.getstr("whitepoint.colortemp.locus.curve")
        )
        self.whitepoint_label.SetMaxFontSize(11)
        self.whitepoint_label.SetForegroundColour(FGCOLOUR)
        self.options_sizer.Add(self.whitepoint_label, flag=wx.ALIGN_CENTER_VERTICAL)
        self.whitepoint_select = wx.Choice(
            self,
            -1,
            size=(150 * scale, -1),
            choices=[
                lang.getstr("calibration.file.none"),
                lang.getstr("whitepoint.colortemp.locus.daylight"),
                lang.getstr("whitepoint.colortemp.locus.blackbody"),
            ],
        )
        self.options_sizer.Add(
            self.whitepoint_select, flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND
        )
        self.whitepoint_select.Bind(wx.EVT_CHOICE, self.generic_select_handler)
        self.whitepoint_select.SetSelection(0)
        self.whitepoint_bmp.Hide()

        # Comparison profile select
        self.comparison_profile_bmp = wx.StaticBitmap(
            self, -1, getbitmap("theme/dashed-16x2-666")
        )
        self.options_sizer.Add(
            self.comparison_profile_bmp, flag=wx.ALIGN_CENTER_VERTICAL
        )
        self.comparison_profile_label = wx.StaticText(
            self, -1, lang.getstr("comparison_profile")
        )
        self.comparison_profile_label.SetMaxFontSize(11)
        self.comparison_profile_label.SetForegroundColour(FGCOLOUR)
        self.options_sizer.Add(
            self.comparison_profile_label, flag=wx.ALIGN_CENTER_VERTICAL
        )
        self.comparison_profiles = dict([(lang.getstr("calibration.file.none"), None)])
        srgb = None
        try:
            srgb = ICCP.ICCProfile(get_data_path("ref/sRGB.icm"))
        except EnvironmentError:
            pass
        except Exception as exception:
            print(exception)
        if srgb:
            self.comparison_profiles[srgb.getDescription()] = srgb
        for profile in config.get_standard_profiles():
            desc = profile.getDescription()
            if desc not in self.comparison_profiles:
                self.comparison_profiles[desc] = profile
        self.comparison_profile_select = wx.Choice(
            self, -1, size=(150 * scale, -1), choices=[]
        )
        self.comparison_profiles_sort()
        self.options_sizer.Add(
            self.comparison_profile_select, flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND
        )
        self.comparison_profile_select.Bind(
            wx.EVT_CHOICE, self.comparison_profile_select_handler
        )
        droptarget = FileDrop(
            self.TopLevelParent,
            {
                ".icc": self.comparison_profile_drop_handler,
                ".icm": self.comparison_profile_drop_handler,
            },
        )
        self.comparison_profile_select.SetDropTarget(droptarget)
        if srgb:
            self.comparison_profile_select.SetSelection(1)
            self.comparison_profile_select.SetToolTipString(srgb.fileName)

        # Rendering intent select
        self.options_sizer.Add((0, 0))
        self.rendering_intent_label = wx.StaticText(
            self, -1, lang.getstr("rendering_intent")
        )
        self.rendering_intent_label.SetMaxFontSize(11)
        self.rendering_intent_label.SetForegroundColour(FGCOLOUR)
        self.options_sizer.Add(
            self.rendering_intent_label, flag=wx.ALIGN_CENTER_VERTICAL
        )
        self.rendering_intent_select = wx.Choice(
            self,
            -1,
            size=(150 * scale, -1),
            choices=[
                lang.getstr("gamap.intents.a"),
                lang.getstr("gamap.intents.r"),
                lang.getstr("gamap.intents.p"),
                lang.getstr("gamap.intents.s"),
            ],
        )
        self.options_sizer.Add(
            self.rendering_intent_select, flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND
        )
        self.rendering_intent_select.Bind(
            wx.EVT_CHOICE, self.rendering_intent_select_handler
        )
        self.rendering_intent_select.SetSelection(0)

        self.options_sizer.Add((0, 0))

        # LUT toggle
        self.toggle_clut = CustomCheckBox(self, -1, "LUT")
        self.toggle_clut.SetForegroundColour(FGCOLOUR)
        self.toggle_clut.SetMaxFontSize(11)
        self.options_sizer.Add(self.toggle_clut, flag=wx.ALIGN_CENTER_VERTICAL)
        self.toggle_clut.Bind(wx.EVT_CHECKBOX, self.toggle_clut_handler)
        self.toggle_clut.Hide()

        # Direction selection
        self.direction_select = wx.Choice(
            self,
            -1,
            size=(150 * scale, -1),
            choices=[
                lang.getstr("direction.forward"),
                lang.getstr("direction.backward.inverted"),
            ],
        )
        self.options_sizer.Add(
            self.direction_select, flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND
        )
        self.direction_select.Bind(wx.EVT_CHOICE, self.direction_select_handler)
        self.direction_select.SetSelection(0)

        self.sizer.Add((0, 0))

    def DrawCanvas(self, profile_no=None, reset=True):
        # Gamut plot
        parent = self.TopLevelParent
        parent.client.SetEnableCenterLines(False)
        parent.client.SetEnableDiagonals(False)
        parent.client.SetEnableGrid(True)
        parent.client.SetEnablePointLabel(False)
        try:
            parent.client.setup(
                [self.comparison_profile, parent.profile],
                profile_no,
                intent=self.intent,
                direction=self.direction,
                order=self.order,
            )
        except Exception as exception:
            show_result_dialog(exception, parent)
        if reset:
            parent.client.reset()
            parent.client.resetzoom()
        wx.CallAfter(self.draw, center=reset)
        wx.CallAfter(parent.handle_errors)

    @property
    def colorspace(self):
        return self.get_colorspace()

    @property
    def comparison_profile(self):
        index = max(self.comparison_profile_select.GetSelection(), 0)
        return list(self.comparison_profiles.values())[index]

    def comparison_profile_drop_handler(self, path):
        try:
            profile = ICCP.ICCProfile(path)
        except Exception as exception:
            show_result_dialog(exception, self.TopLevelParent)
        else:
            desc = profile.getDescription()
            self.comparison_profiles[desc] = profile
            self.comparison_profiles_sort()
            self.comparison_profile_select.SetSelection(
                list(self.comparison_profiles.keys()).index(desc)
            )
            self.comparison_profile_select_handler(None)

    def comparison_profile_select_handler(self, event):
        try:
            index = self.comparison_profile_select.GetSelection()
            if index > 0:
                self.comparison_profile_select.SetToolTipString(
                    self.comparison_profile.fileName
                )
            else:
                self.comparison_profile_select.SetToolTip(None)
            self.comparison_whitepoint_bmp.Show(index > 0)
            self.comparison_whitepoint_legend.Show(index > 0)
            self.comparison_profile_bmp.Show(index > 0)
            self.DrawCanvas(0, reset=False)
        except Exception:
            traceback.print_exc()

    def comparison_profiles_sort(self):
        comparison_profiles = dict_slice(self.comparison_profiles, 2)
        comparison_profiles = dict_sort(comparison_profiles, key=lambda s: s.lower())
        self.comparison_profiles = dict_slice(self.comparison_profiles, None, 2)
        self.comparison_profiles.update(comparison_profiles)
        self.comparison_profile_select.SetItems(list(self.comparison_profiles.keys()))

    @property
    def direction(self):
        if self.direction_select.IsShown():
            return {0: "f", 1: "ib"}.get(self.direction_select.GetSelection())
        else:
            return "f"

    def draw(self, center=False):
        colorspace = self.colorspace
        parent = self.TopLevelParent
        parent.client.proportional = True
        parent.client.DrawCanvas(
            f"{colorspace} {lang.getstr('colorspace')}",
            colorspace,
            whitepoint=self.whitepoint_select.GetSelection(),
            center=center,
            show_outline=self.draw_gamut_outline_cb.GetValue(),
        )

    def draw_gamut_outline_handler(self, event):
        self.colorspace_outline_bmp.Show(self.draw_gamut_outline_cb.GetValue())
        self.draw()

    def generic_select_handler(self, event):
        self.whitepoint_bmp.Show(self.whitepoint_select.GetSelection() > 0)
        parent = self.TopLevelParent
        if parent.client.profiles:
            self.draw(center=event.GetId() == self.colorspace_select.GetId())
        else:
            self.DrawCanvas()

    def get_colorspace(self, dimensions=2):
        return {
            0: "a*b*" if dimensions == 2 else "Lab",
            1: "u*v*" if dimensions == 2 else "Luv",
            2: "u'v'" if dimensions == 2 else "Lu'v'",
            3: "xy" if dimensions == 2 else "xyY",
            4: "DIN99",
            5: "DIN99b",
            6: "DIN99c",
            7: "DIN99d",
            8: "ICtCp",
            9: "IPT",
            10: "Lpt",
        }.get(
            self.colorspace_select.GetSelection(), "a*b*" if dimensions == 2 else "Lab"
        )

    @property
    def intent(self):
        return {0: "a", 1: "r", 2: "p", 3: "s"}.get(
            self.rendering_intent_select.GetSelection()
        )

    @property
    def order(self):
        parent = self.TopLevelParent
        return {True: "n", False: "r"}.get(
            bool(parent.profile)
            and not ("B2A0" in parent.profile.tags or "A2B0" in parent.profile.tags)
            or self.toggle_clut.GetValue()
        )

    def rendering_intent_select_handler(self, event):
        self.DrawCanvas(reset=False)

    def direction_select_handler(self, event):
        self.DrawCanvas(reset=False)

    def toggle_clut_handler(self, event):
        parent = self.TopLevelParent
        self.Freeze()
        self.direction_select.Enable(
            bool(parent.profile)
            and "B2A0" in parent.profile.tags
            and "A2B0" in parent.profile.tags
            and self.toggle_clut.GetValue()
        )
        self.DrawCanvas(reset=False)
        self.Thaw()


class PIFrame_2WaySplitter(TwoWaySplitter):
    def OnLeftDClick(self, event):
        if not self.IsEnabled():
            return

        pt = event.GetPosition()

        if self.GetMode(pt):

            barSize = self._GetSashSize()

            winborder, titlebar = get_platform_window_decoration_size()

            win0w = self.GetTopLeft().Size[0]

            self.Freeze()
            TwoWaySplitter.OnLeftDClick(self, event)
            if self._expanded < 0:
                self.Parent.SetMinSize(
                    (
                        defaults["size.profile_info.split.w"] + winborder * 2,
                        self.Parent.GetMinSize()[1],
                    )
                )
                if (
                    not self.Parent.IsMaximized()
                    and self.Parent.ClientSize[0] < self.GetSplitSize()[0]
                ):
                    self.Parent.ClientSize = (
                        self.GetSplitSize()[0],
                        self.Parent.ClientSize[1],
                    )
            else:
                self.Parent.SetMinSize(
                    (
                        defaults["size.profile_info.w"] + winborder * 2,
                        self.Parent.GetMinSize()[1],
                    )
                )
                w = max(self.GetExpandedSize()[0], self._splitx)
                if not self.Parent.IsMaximized() and self.Parent.ClientSize[0] > w:
                    self.Parent.ClientSize = (w, self.Parent.ClientSize[1])
            if os.getenv("XDG_SESSION_TYPE") == "wayland":
                self.Parent.MaxSize = self.Parent.Size
                wx.CallAfter(set_maxsize, self.Parent, (-1, -1))
            if self.GetTopLeft().Size[0] != win0w:
                self.Parent.redraw()
            self.Parent.resize_grid()
            self.Thaw()


class ProfileInfoFrame(LUTFrame):
    def __init__(self, *args, **kwargs):

        if len(args) < 3 and "title" not in kwargs:
            kwargs["title"] = lang.getstr("profile.info")
        if "name" not in kwargs:
            kwargs["name"] = "profile_info"

        BaseFrame.__init__(self, *args, **kwargs)

        self.SetIcons(
            config.get_icon_bundle([256, 48, 32, 16], f"{appname}-profile-info")
        )

        self.profile = None
        self.xLabel = lang.getstr("in")
        self.yLabel = lang.getstr("out")

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)

        self.splitter = PIFrame_2WaySplitter(
            self, -1, agwStyle=wx.SP_LIVE_UPDATE | wx.SP_NOSASH
        )
        self.sizer.Add(self.splitter, 1, flag=wx.EXPAND)

        p1 = wx_Panel(self.splitter, name="canvaspanel")
        p1.SetBackgroundColour(BGCOLOUR)
        p1.sizer = wx.BoxSizer(wx.VERTICAL)
        p1.SetSizer(p1.sizer)
        self.splitter.AppendWindow(p1)

        self.plot_mode_sizer = wx.BoxSizer(wx.HORIZONTAL)
        p1.sizer.Add(self.plot_mode_sizer, flag=wx.ALIGN_CENTER | wx.TOP, border=12)

        # The order of the choice items is significant for compatibility
        # with LUTFrame:
        # 0 = vcgt,
        # 1 = [rgb]TRC
        # 2 = gamut
        self.plot_mode_select = wx.Choice(
            p1,
            -1,
            choices=[
                lang.getstr("vcgt"),
                lang.getstr("[rgb]TRC"),
                lang.getstr("gamut"),
            ],
        )
        self.plot_mode_sizer.Add(self.plot_mode_select, flag=wx.ALIGN_CENTER_VERTICAL)
        self.plot_mode_select.Bind(wx.EVT_CHOICE, self.plot_mode_select_handler)
        self.plot_mode_select.Disable()

        self.tooltip_btn = BitmapButton(
            p1, -1, geticon(16, "question-inverted"), style=wx.NO_BORDER
        )
        self.tooltip_btn.SetBackgroundColour(BGCOLOUR)
        self.tooltip_btn.Bind(wx.EVT_BUTTON, self.tooltip_handler)
        self.tooltip_btn.SetToolTipString(lang.getstr("gamut_plot.tooltip"))
        self.plot_mode_sizer.Add(
            self.tooltip_btn, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=8
        )

        self.save_plot_btn = BitmapButton(
            p1, -1, geticon(16, "image-x-generic-inverted"), style=wx.NO_BORDER
        )
        self.save_plot_btn.SetBackgroundColour(BGCOLOUR)
        self.save_plot_btn.Bind(wx.EVT_BUTTON, self.SaveFile)
        self.save_plot_btn.SetToolTipString(
            f"{lang.getstr('save_as')} (*.bmp, *.xbm, *.xpm, *.jpg, *.png)"
        )
        self.save_plot_btn.Disable()
        self.plot_mode_sizer.Add(
            self.save_plot_btn, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=8
        )

        self.view_3d_btn = BitmapButton(p1, -1, geticon(16, "3D"), style=wx.NO_BORDER)
        self.view_3d_btn.SetBackgroundColour(BGCOLOUR)
        self.view_3d_btn.Bind(wx.EVT_BUTTON, self.view_3d)
        self.view_3d_btn.Bind(wx.EVT_CONTEXT_MENU, self.view_3d_format_popup)
        self.view_3d_btn.SetToolTipString(lang.getstr("view.3d"))
        self.view_3d_btn.Disable()
        self.plot_mode_sizer.Add(
            self.view_3d_btn, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=12
        )
        self.view_3d_format_btn = BitmapButton(
            p1, -1, getbitmap("theme/dropdown-arrow"), style=wx.NO_BORDER
        )
        self.view_3d_format_btn.MinSize = (-1, self.view_3d_btn.Size[1])
        self.view_3d_format_btn.SetBackgroundColour(BGCOLOUR)
        self.view_3d_format_btn.Bind(wx.EVT_BUTTON, self.view_3d_format_popup)
        self.view_3d_format_btn.Bind(wx.EVT_CONTEXT_MENU, self.view_3d_format_popup)
        self.view_3d_format_btn.SetToolTipString(lang.getstr("view.3d"))
        self.view_3d_format_btn.Disable()
        self.plot_mode_sizer.Add(
            self.view_3d_format_btn, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=4
        )

        self.client = GamutCanvas(p1)
        p1.sizer.Add(
            self.client,
            1,
            flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP | wx.BOTTOM,
            border=4,
        )

        self.options_panel = SimpleBook(p1)
        self.options_panel.SetBackgroundColour(BGCOLOUR)
        p1.sizer.Add(self.options_panel, flag=wx.EXPAND | wx.BOTTOM, border=8)

        self.status = BitmapBackgroundPanelText(p1)
        self.status.SetMaxFontSize(11)
        self.status.label_y = 0
        self.status.textshadow = False
        self.status.SetBackgroundColour(BGCOLOUR)
        self.status.SetForegroundColour(FGCOLOUR)
        h = self.status.GetTextExtent("Ig")[1]
        self.status.SetMinSize((0, h * 2 + 10))
        p1.sizer.Add(self.status, flag=wx.EXPAND)

        self.gamut_status = BitmapBackgroundPanelTextGamut(p1)
        self.gamut_status.SetMaxFontSize(11)
        self.gamut_status.label_y = 0
        self.gamut_status.textshadow = False
        self.gamut_status.SetBackgroundColour(BGCOLOUR)
        self.gamut_status.SetForegroundColour(FGCOLOUR)
        h = self.gamut_status.GetTextExtent("Ig")[1]
        self.gamut_status.SetMinSize((0, h * 2 + 10))
        p1.sizer.Add(self.gamut_status, flag=wx.EXPAND)

        # Gamut view options
        self.gamut_view_options = GamutViewOptions(p1)
        self.options_panel.AddPage(self.gamut_view_options, "")

        # Curve view options
        self.lut_view_options = wx_Panel(p1, name="lut_view_options")
        self.lut_view_options.SetBackgroundColour(BGCOLOUR)
        self.lut_view_options_sizer = self.box_sizer = wx.FlexGridSizer(0, 3, 4, 4)
        self.lut_view_options_sizer.AddGrowableCol(0)
        self.lut_view_options_sizer.AddGrowableCol(2)
        self.lut_view_options.SetSizer(self.lut_view_options_sizer)
        self.options_panel.AddPage(self.lut_view_options, "")

        self.lut_view_options_sizer.Add((0, 0))

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        self.lut_view_options_sizer.Add(
            hsizer, flag=wx.ALIGN_CENTER | wx.BOTTOM, border=8
        )

        self.rendering_intent_select = wx.Choice(
            self.lut_view_options,
            -1,
            choices=[
                lang.getstr("gamap.intents.a"),
                lang.getstr("gamap.intents.r"),
                lang.getstr("gamap.intents.p"),
                lang.getstr("gamap.intents.s"),
            ],
        )
        hsizer.Add((10, self.rendering_intent_select.Size[1]))
        hsizer.Add(self.rendering_intent_select, flag=wx.ALIGN_CENTER_VERTICAL)
        self.rendering_intent_select.Bind(
            wx.EVT_CHOICE, self.rendering_intent_select_handler
        )
        self.rendering_intent_select.SetSelection(1)

        self.direction_select = wx.Choice(
            self.lut_view_options,
            -1,
            choices=[
                lang.getstr("direction.backward"),
                lang.getstr("direction.forward.inverted"),
            ],
        )
        hsizer.Add(
            self.direction_select, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=10
        )
        self.direction_select.Bind(wx.EVT_CHOICE, self.direction_select_handler)
        self.direction_select.SetSelection(0)

        self.lut_view_options_sizer.Add((0, 0))

        self.lut_view_options_sizer.Add((0, 0))

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        self.lut_view_options_sizer.Add(hsizer, flag=wx.ALIGN_CENTER)

        hsizer.Add((16, 0))

        self.show_as_L = CustomCheckBox(self.lut_view_options, -1, "L* \u2192")
        self.show_as_L.SetForegroundColour(FGCOLOUR)
        self.show_as_L.SetValue(True)
        hsizer.Add(self.show_as_L, flag=wx.ALIGN_CENTER_VERTICAL | wx.RIGHT, border=4)
        self.show_as_L.Bind(wx.EVT_CHECKBOX, self.DrawLUT)

        self.add_toggles(self.lut_view_options, hsizer)

        self.toggle_clut = CustomCheckBox(self.lut_view_options, -1, "LUT")
        self.toggle_clut.SetForegroundColour(FGCOLOUR)
        hsizer.Add(self.toggle_clut, flag=wx.ALIGN_CENTER_VERTICAL | wx.LEFT, border=16)
        self.toggle_clut.Bind(wx.EVT_CHECKBOX, self.toggle_clut_handler)

        self.lut_view_options_sizer.Add((0, 0))

        p2 = wx_Panel(self.splitter, name="gridpanel")
        p2.sizer = wx.BoxSizer(wx.VERTICAL)
        p2.SetSizer(p2.sizer)
        self.splitter.AppendWindow(p2)

        self.grid = CustomGrid(p2, -1)
        self.grid.alternate_row_label_background_color = wx.Colour(230, 230, 230)
        self.grid.alternate_cell_background_color = (
            self.grid.alternate_row_label_background_color
        )
        self.grid.draw_horizontal_grid_lines = False
        self.grid.draw_vertical_grid_lines = False
        self.grid.draw_row_labels = False
        self.grid.show_cursor_outline = False
        self.grid.style = ""
        self.grid.CreateGrid(0, 2)
        self.grid.SetCellHighlightPenWidth(0)
        self.grid.SetCellHighlightROPenWidth(0)
        self.grid.SetDefaultCellBackgroundColour(self.grid.GetLabelBackgroundColour())
        font = self.grid.GetDefaultCellFont()
        if font.PointSize > 11:
            font.PointSize = 11
            self.grid.SetDefaultCellFont(font)
        self.grid.SetSelectionMode(wx.grid.Grid.wxGridSelectRows)
        self.grid.SetRowLabelSize(self.grid.GetDefaultRowSize())
        self.grid.SetColLabelSize(0)
        self.grid.DisableDragRowSize()
        self.grid.EnableDragColSize()
        self.grid.EnableEditing(False)
        self.grid.EnableGridLines(False)
        self.grid.Bind(wx.grid.EVT_GRID_COL_SIZE, self.resize_grid)
        p2.sizer.Add(self.grid, 1, flag=wx.EXPAND)

        drophandlers = {".icc": self.drop_handler, ".icm": self.drop_handler}
        self.droptarget = FileDrop(self, drophandlers)
        self.client.SetDropTarget(self.droptarget)
        droptarget = FileDrop(self, drophandlers)
        self.grid.SetDropTarget(droptarget)

        min_w = max(
            self.options_panel.GetBestSize()[0] + 24,
            defaults["size.profile_info.w"] - self.splitter._GetSashSize(),
        )

        self.splitter.SetMinimumPaneSize(min_w)
        self.splitter.SetHSplit(0)

        border, titlebar = get_platform_window_decoration_size()
        self.SetSaneGeometry(
            getcfg("position.profile_info.x"),
            getcfg("position.profile_info.y"),
            defaults["size.profile_info.split.w"],
            getcfg("size.profile_info.h"),
        )
        self.SetMinSize(
            (min_w + border * 2, defaults["size.profile_info.h"] + titlebar + border)
        )
        self.splitter.SetSplitSize(
            (defaults["size.profile_info.split.w"], self.ClientSize[1])
        )
        self.splitter.SetExpandedSize(self.ClientSize)

        self.client.canvas.Bind(wx.EVT_MOTION, self.OnMotion)
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.Bind(wx.EVT_MOVE, self.OnMove)
        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_SPLITTER_SASH_POS_CHANGING, self.OnSashPosChanging)

        self.display_no = -1
        self.display_rects = get_display_rects()

        children = self.GetAllChildren()

        self.Bind(wx.EVT_KEY_DOWN, self.key_handler)
        for child in children:
            if isinstance(child, wx.Choice):
                child.SetMaxFontSize(11)
            child.Bind(wx.EVT_KEY_DOWN, self.key_handler)
            child.Bind(wx.EVT_MOUSEWHEEL, self.OnWheel)
            if (
                sys.platform == "win32"
                and sys.getwindowsversion() >= (6,)
                and isinstance(child, wx.Panel)
            ):
                # No need to enable double buffering under Linux and Mac OS X.
                # Under Windows, enabling double buffering on the panel seems
                # to work best to reduce flicker.
                child.SetDoubleBuffered(True)

    def DrawCanvas(self, profile_no=None, reset=True):
        self.SetStatusText("")
        if self.plot_mode_select.GetSelection() == self.plot_mode_select.GetCount() - 1:
            # Gamut plot
            self.plot_mode_sizer.Show(self.view_3d_btn)
            self.plot_mode_sizer.Show(self.view_3d_format_btn)
            self.options_panel.GetCurrentPage().DrawCanvas(reset=reset)
            self.save_plot_btn.Enable()
            self.view_3d_btn.Enable()
            self.view_3d_format_btn.Enable()
        else:
            # Curves plot
            self.plot_mode_sizer.Hide(self.view_3d_btn)
            self.plot_mode_sizer.Hide(self.view_3d_format_btn)
            self.client.SetEnableCenterLines(True)
            self.client.SetEnableDiagonals("Bottomleft-Topright")
            self.client.SetEnableGrid(False)
            self.client.SetEnablePointLabel(True)
            if (
                isinstance(self.profile.tags.get("vcgt"), ICCP.VideoCardGammaType)
                or (self.rTRC and self.gTRC and self.bTRC)
                or ("B2A0" in self.profile.tags or "A2B0" in self.profile.tags)
            ):
                self.DrawLUT()
                self.handle_errors()
            else:
                self.client.DrawLUT()
        self.splitter.GetTopLeft().sizer.Layout()
        self.splitter.GetTopLeft().Refresh()

    def LoadProfile(self, profile, reset=True):
        if not isinstance(profile, ICCP.ICCProfile):
            try:
                profile = ICCP.ICCProfile(profile)
            except (IOError, ICCP.ICCProfileInvalidError):
                show_result_dialog(
                    Error(f"{lang.getstr('profile.invalid')}\n{profile}"), self
                )
                self.DrawCanvas(reset=reset)
                return
        if (
            not reset
            and getattr(self, "profile", None)
            and self.profile.colorSpace != profile.colorSpace
        ):
            reset = True
        self.profile = profile
        cinfo = []
        vinfo = []
        if "meta" in profile.tags:
            for key in ("avg", "max", "rms"):
                try:
                    dE = float(profile.tags.meta.getvalue(f"ACCURACY_dE76_{key}"))
                except (TypeError, ValueError):
                    pass

            gamuts = (
                ("srgb", "sRGB", ICCP.GAMUT_VOLUME_SRGB),
                ("adobe-rgb", "Adobe RGB", ICCP.GAMUT_VOLUME_ADOBERGB),
                ("dci-p3", "DCI P3", ICCP.GAMUT_VOLUME_SMPTE431_P3),
            )
            for key, name, _volume in gamuts:
                try:
                    gamut_coverage = float(
                        profile.tags.meta.getvalue(f"GAMUT_coverage({key})")
                    )
                except (TypeError, ValueError):
                    traceback.print_exc()
                    gamut_coverage = None
                if gamut_coverage:
                    cinfo.append(f"{gamut_coverage * 100:.1f}% {name}")
            try:
                gamut_volume = float(profile.tags.meta.getvalue("GAMUT_volume"))
            except (TypeError, ValueError):
                traceback.print_exc()
                gamut_volume = None
            if gamut_volume:
                for _key, name, volume in gamuts:
                    vinfo.append(
                        f"{gamut_volume * ICCP.GAMUT_VOLUME_SRGB / volume * 100:.1f}% {name}"
                    )
                    if len(vinfo) == len(cinfo):
                        break

            if gamut_coverage and gamut_volume:
                gamut_cov_text = f"{cinfo[0]}    {cinfo[1]}    {cinfo[2]}"
                self.SetGamutStatusText(gamut_cov_text)

        for channel in "rgb":
            trc = profile.tags.get(f"{channel}TRC", profile.tags.get("kTRC"))
            if isinstance(trc, ICCP.ParametricCurveType):
                trc = trc.get_trc()
            setattr(self, f"{channel}TRC", trc)
            setattr(self, f"tf_{channel}TRC", trc)
        self.trc = None

        self.gamut_view_options.direction_select.Enable(
            "B2A0" in self.profile.tags and "A2B0" in self.profile.tags
        )
        if not self.gamut_view_options.direction_select.Enabled:
            self.gamut_view_options.direction_select.SetSelection(0)
        self.gamut_view_options.toggle_clut.SetValue(
            "B2A0" in profile.tags or "A2B0" in profile.tags
        )
        self.gamut_view_options.toggle_clut.Show(
            "B2A0" in profile.tags or "A2B0" in profile.tags
        )
        self.gamut_view_options.toggle_clut.Enable(
            isinstance(self.rTRC, ICCP.CurveType)
            and isinstance(self.gTRC, ICCP.CurveType)
            and isinstance(self.bTRC, ICCP.CurveType)
        )
        self.toggle_clut.SetValue("B2A0" in profile.tags or "A2B0" in profile.tags)

        plot_mode = self.plot_mode_select.GetStringSelection()
        choice = []
        info = profile.get_info()
        self.client.errors = []
        if (self.rTRC and self.gTRC and self.bTRC) or (
            self.toggle_clut.GetValue()
            and profile.colorSpace in (b"RGB", b"GRAY", b"CMYK")
        ):
            if "vcgt" in profile.tags or "MS00" in profile.tags:
                choice.append(lang.getstr("vcgt"))
            try:
                self.lookup_tone_response_curves()
            except Exception as exception:
                wx.CallAfter(show_result_dialog, exception, self)
            else:
                choice.append(lang.getstr("[rgb]TRC"))
        choice = self.add_shaper_curves(choice)
        choice.append(lang.getstr("gamut"))
        self.Freeze()
        self.plot_mode_select.SetItems(choice)
        if not self.plot_mode_select.SetStringSelection(plot_mode):
            self.plot_mode_select.SetSelection(self.plot_mode_select.GetCount() - 1)
            reset = True
        self.select_current_page()
        self.plot_mode_select.Enable()

        self.SetTitle(
            " \u2014 ".join([lang.getstr("profile.info"), profile.getDescription()])
        )

        rows = [("", "")]
        lines = []
        for label, value in info:
            label = label.replace("\0", "")
            if value is None:
                value = ""
            elif not isinstance(value, str):
                value = str(value, "ascii")
            value = (
                wrap(universal_newlines(value.strip()).replace("\t", "\n"), 52)
                .replace("\0", "")
                .split("\n")
            )
            linecount = len(value)
            for i, line in enumerate(value):
                value[i] = line.strip()
            label = universal_newlines(label).split("\n")
            while len(label) < linecount:
                label.append("")
            lines.extend(list(zip(label, value)))
        for i, line in enumerate(lines):
            line = list(line)
            indent = re.match(r"\s+", line[0])
            for j, v in enumerate(line):
                if v.endswith("_"):
                    continue
                key = re.sub(
                    r"[^0-9A-Za-z]+",
                    "_",
                    strtr(line[j], {"\u0394E": "Delta E"}).lower().strip(),
                    0,
                ).strip("_")
                val = lang.getstr(key)
                if key != val:
                    line[j] = val
            if indent:
                # if i + 1 < len(lines) and lines[i + 1][0].startswith(" "):
                # marker = u" ├  "
                # else:
                # marker = u" └  "
                line[0] = indent.group() + line[0].strip()
            elif line[0]:
                # if i + 1 < len(lines) and lines[i + 1][0].startswith(" "):
                # marker = u"▼ "
                # else:
                # marker = u"► "
                pass
            rows.append(line)

        rows.append(("", ""))

        if self.grid.GetNumberRows():
            self.grid.DeleteRows(0, self.grid.GetNumberRows())
        self.grid.AppendRows(len(rows))
        labelcolor = self.grid.GetLabelTextColour()
        alpha = 102
        namedcolor = False
        for i, (label, value) in enumerate(rows):
            self.grid.SetCellValue(i, 0, f" {label}")
            bgcolor = self.grid.GetCellBackgroundColour(i, 0)
            bgblend = (255 - alpha) / 255.0
            blend = alpha / 255.0
            textcolor = wx.Colour(
                int(round(bgblend * bgcolor.Red() + blend * labelcolor.Red())),
                int(round(bgblend * bgcolor.Green() + blend * labelcolor.Green())),
                int(round(bgblend * bgcolor.Blue() + blend * labelcolor.Blue())),
            )
            if label == lang.getstr("named_colors"):
                namedcolor = True
            elif label.strip() and label.lstrip() == label:
                namedcolor = False
            if namedcolor:
                color = re.match(r"(Lab|XYZ)((?:\s+-?\d+(?:\.\d+)){3,})", value)
            else:
                color = None
            if color:
                if color.groups()[0] == "Lab":
                    color = colormath.Lab2RGB(
                        *[float(v) for v in color.groups()[1].strip().split()],
                        **dict(scale=255),
                    )
                else:
                    # XYZ
                    color = colormath.XYZ2RGB(
                        *[float(v) for v in color.groups()[1].strip().split()],
                        **dict(scale=255),
                    )
                labelbgcolor = wx.Colour(*[int(round(v)) for v in color])
                rowlabelrenderer = CustomRowLabelRenderer(labelbgcolor)
            else:
                rowlabelrenderer = None
            self.grid.SetRowLabelRenderer(i, rowlabelrenderer)
            self.grid.SetCellTextColour(i, 0, textcolor)
            self.grid.SetCellValue(i, 1, value)

        self.grid.AutoSizeColumn(0)
        self.resize_grid()
        self.grid.ClearSelection()
        self.Layout()
        self.DrawCanvas(reset=reset)
        self.Thaw()

    def OnClose(self, event):
        if getattr(self.worker, "thread", None) and self.worker.thread.is_alive():
            self.worker.abort_subprocess(True)
            return
        self.worker.wrapup(False)
        writecfg(
            module="profile-info",
            options=(
                "3d.format",
                "last_cal_or_icc_path",
                "last_icc_path",
                "position.profile_info",
                "size.profile_info",
            ),
        )
        # Hide first (looks nicer)
        self.Hide()
        # Need to use CallAfter to prevent hang under Windows if minimized
        wx.CallAfter(self.Destroy)

    def OnMotion(self, event):
        if isinstance(event, wx.MouseEvent):
            xy = self.client._getXY(event)
        else:
            xy = event
        if self.plot_mode_select.GetSelection() < self.plot_mode_select.GetCount() - 1:
            # Curves plot
            if isinstance(event, wx.MouseEvent):
                if not event.LeftIsDown():
                    self.UpdatePointLabel(xy)
                else:
                    self.client.erase_pointlabel()
        else:
            # Gamut plot
            page = self.options_panel.GetCurrentPage()
            if page.colorspace in ("a*b*", "u*v*") or page.colorspace.startswith(
                "DIN99"
            ):
                format_ = "%.2f %.2f"
            else:
                format_ = "%.4f %.4f"
            whitepoint_no = page.whitepoint_select.GetSelection()
            if whitepoint_no > 0:
                if page.colorspace == "a*b*":
                    x, y, Y = colormath.Lab2xyY(100.0, xy[0], xy[1])
                elif page.colorspace == "u*v*":
                    X, Y, Z = colormath.Luv2XYZ(100.0, xy[0], xy[1])
                    x, y = colormath.XYZ2xyY(X, Y, Z)[:2]
                elif page.colorspace == "u'v'":
                    x, y = colormath.u_v_2xy(xy[0], xy[1])
                elif page.colorspace.startswith("DIN99"):
                    # DIN99
                    L99 = colormath.Lab2DIN99(100, 0, 0)[0]
                    if page.colorspace == "DIN99b":
                        # DIN99b
                        a, b = colormath.DIN99b2Lab(L99, xy[0], xy[1])[1:]
                    elif page.colorspace == "DIN99c":
                        # DIN99c
                        a, b = colormath.DIN99c2Lab(L99, xy[0], xy[1])[1:]
                    elif page.colorspace == "DIN99d":
                        # DIN99d
                        a, b = colormath.DIN99d2Lab(L99, xy[0], xy[1])[1:]
                    else:
                        # DIN99
                        a, b = colormath.DIN992Lab(L99, xy[0], xy[1])[1:]
                    x, y = colormath.Lab2xyY(100.0, a, b)[:2]
                elif page.colorspace == "ICtCp":
                    X, Y, Z = colormath.ICtCp2XYZ(1.0, xy[0], xy[1])
                    x, y = colormath.XYZ2xyY(X, Y, Z)[:2]
                elif page.colorspace == "IPT":
                    X, Y, Z = colormath.IPT2XYZ(1.0, xy[0], xy[1])
                    x, y = colormath.XYZ2xyY(X, Y, Z)[:2]
                elif page.colorspace == "Lpt":
                    X, Y, Z = colormath.Lpt2XYZ(100.0, xy[0], xy[1])
                    x, y = colormath.XYZ2xyY(X, Y, Z)[:2]
                else:
                    # xy
                    x, y = xy
                cct, delta = colormath.xy_CCT_delta(x, y, daylight=whitepoint_no == 1)
                status = format_ % xy
                if cct:
                    if delta:
                        locus = {"Blackbody": "blackbody", "Daylight": "daylight"}.get(
                            page.whitepoint_select.GetStringSelection(),
                            page.whitepoint_select.GetStringSelection(),
                        )
                        status = "%s, CCT %i (%s %.2f)" % (
                            format_ % xy,
                            cct,
                            lang.getstr("delta_e_to_locus", locus),
                            delta["E"],
                        )
                    else:
                        status = "%s, CCT %i" % (format_ % xy, cct)
                self.SetStatusText(status)
            else:
                self.SetStatusText(format_ % xy)
        if isinstance(event, wx.MouseEvent):
            event.Skip()  # Go to next handler

    def OnMove(self, event=None):
        if self.IsShownOnScreen() and not self.IsMaximized() and not self.IsIconized():
            x, y = self.GetScreenPosition()
            setcfg("position.profile_info.x", x)
            setcfg("position.profile_info.y", y)
        if event:
            event.Skip()

    def OnSashPosChanging(self, event):
        self.splitter.SetExpandedSize(
            (self.splitter._splitx + self.splitter._GetSashSize(), self.ClientSize[1])
        )
        setcfg(
            "size.profile_info.w", self.splitter._splitx + self.splitter._GetSashSize()
        )
        wx.CallAfter(self.redraw)
        wx.CallAfter(self.resize_grid)

    def OnSize(self, event=None):
        if (
            self.IsShownOnScreen()
            and self.IsMaximized()
            and self.splitter._expanded < 0
        ):
            self.splitter.AdjustLayout()
        self.splitter.Refresh()
        wx.CallAfter(self.redraw)
        wx.CallAfter(self.resize_grid)
        if self.IsShownOnScreen() and not self.IsIconized():
            wx.CallAfter(self._setsize)
        if event:
            event.Skip()

    def OnWheel(self, event):
        xy = wx.GetMousePosition()
        if not self.client.GetClientRect().Contains(self.client.ScreenToClient(xy)):
            if self.grid.GetClientRect().Contains(self.grid.ScreenToClient(xy)):
                self.grid.SetFocus()
            event.Skip()
        elif self.client.last_draw:
            if event.WheelRotation < 0:
                direction = 1.0
            else:
                direction = -1.0
            self.client.zoom(direction)

    def _setsize(self):
        if not self.IsMaximized():
            w, h = self.ClientSize
            if self.splitter._expanded < 0:
                self.splitter.SetExpandedSize(
                    (self.splitter._splitx + self.splitter._GetSashSize(), h)
                )
                self.splitter.SetSplitSize((w, h))
                setcfg(
                    "size.profile_info.w",
                    self.splitter._splitx + self.splitter._GetSashSize(),
                )
                setcfg("size.profile_info.split.w", w)
            else:
                self.splitter.SetExpandedSize((w, h))
                setcfg("size.profile_info.w", w)
            setcfg("size.profile_info.h", h)

    def drop_handler(self, path):
        """Drag'n'drop handler for .cal/.icc/.icm files."""
        reset = (
            self.plot_mode_select.GetSelection() == self.plot_mode_select.GetCount() - 1
        )
        self.LoadProfile(path, reset=reset)

    def get_platform_window_size(
        self, defaultwidth=None, defaultheight=None, split=False
    ):
        if split:
            name = ".split"
        else:
            name = ""
        if not defaultwidth:
            defaultwidth = defaults[f"size.profile_info{name}.w"]
        if not defaultheight:
            defaultheight = defaults["size.profile_info.h"]
        border, titlebar = get_platform_window_decoration_size()
        # return (max(max(getcfg("size.profile_info{name}.w"),
        # defaultwidth), self.GetMinSize()[0] - border * 2),
        # max(getcfg("size.profile_info.h"),
        # defaultheight))
        if split:
            w, h = self.splitter.GetSplitSize()
        else:
            w, h = self.splitter.GetExpandedSize()
        return (
            max(w, defaultwidth, self.GetMinSize()[0] - border * 2),
            max(h, defaultheight),
        )

    def key_handler(self, event):
        # AltDown
        # CmdDown
        # ControlDown
        # GetKeyCode
        # GetModifiers
        # GetPosition
        # GetRawKeyCode
        # GetRawKeyFlags
        # GetUniChar
        # GetUnicodeKey
        # GetX
        # GetY
        # HasModifiers
        # KeyCode
        # MetaDown
        # Modifiers
        # Position
        # RawKeyCode
        # RawKeyFlags
        # ShiftDown
        # UnicodeKey
        # X
        # Y
        key = event.GetKeyCode()
        if (
            event.ControlDown() or event.CmdDown()
        ):  # CTRL (Linux/Mac/Windows) / CMD (Mac)
            focus = self.FindFocus()
            if focus and self.grid in (
                focus,
                focus.GetParent(),
                focus.GetGrandParent(),
            ):
                event.Skip()
            if key == 83 and self.profile:  # S
                self.SaveFile()
                return
            else:
                event.Skip()
        elif key in (43, wx.WXK_NUMPAD_ADD):
            # + key zoom in
            self.client.zoom(-1)
        elif key in (45, wx.WXK_NUMPAD_SUBTRACT):
            # - key zoom out
            self.client.zoom(1)
        else:
            event.Skip()

    def plot_mode_select_handler(self, event):
        self.Freeze()
        self.select_current_page()
        reset = (
            self.plot_mode_select.GetSelection() == self.plot_mode_select.GetCount() - 1
        )
        self.DrawCanvas(reset=reset)
        self.Thaw()

    def get_commands(self):
        return self.get_common_commands() + [
            "profile-info [filename]",
            "load <filename>",
        ]

    def process_data(self, data):
        if (data[0] == "profile-info" and len(data) < 3) or (
            data[0] == "load" and len(data) == 2
        ):
            if self.IsIconized():
                self.Restore()
            self.Raise()
            if len(data) == 2:
                path = data[1]
                if not os.path.isfile(path) and not os.path.isabs(path):
                    path = get_data_path(path)
                if not path:
                    return "fail"
                else:
                    self.droptarget.OnDropFiles(0, 0, [path])
            return "ok"
        return "invalid"

    def redraw(self):
        if (
            self.plot_mode_select.GetSelection() == self.plot_mode_select.GetCount() - 1
            and self.client.last_draw
        ):
            # Update gamut plot
            self.client.Parent.Freeze()
            self.client._DrawCanvas(self.client.last_draw[0])
            self.client.Parent.Thaw()

    def resize_grid(self, event=None):
        self.grid.SetColSize(
            1,
            max(
                self.grid.GetSize()[0]
                - self.grid.GetRowLabelSize()
                - self.grid.GetColSize(0)
                - wx.SystemSettings_GetMetric(wx.SYS_VSCROLL_X),
                0,
            ),
        )
        self.grid.SetMargins(
            0 - wx.SystemSettings_GetMetric(wx.SYS_VSCROLL_X),
            0 - wx.SystemSettings_GetMetric(wx.SYS_HSCROLL_Y),
        )
        self.grid.ForceRefresh()

    def select_current_page(self):
        if self.plot_mode_select.GetSelection() == self.plot_mode_select.GetCount() - 1:
            # Gamut plot
            self.options_panel.SetSelection(0)
        else:
            # Curves plot
            self.options_panel.SetSelection(1)
        self.splitter.GetTopLeft().Layout()

    def view_3d(self, event):
        profile = self.profile
        if not profile:
            defaultDir, defaultFile = get_verified_path("last_icc_path")
            dlg = wx.FileDialog(
                self,
                lang.getstr("profile.choose"),
                defaultDir=defaultDir,
                defaultFile=defaultFile,
                wildcard=f'{lang.getstr("filetype.icc")}|*.icc;*.icm',
                style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST,
            )
            dlg.Center(wx.BOTH)
            result = dlg.ShowModal()
            path = dlg.GetPath()
            dlg.Destroy()
            if result != wx.ID_OK:
                return
            try:
                profile = ICCP.ICCProfile(path)
            except (IOError, ICCP.ICCProfileInvalidError):
                show_result_dialog(
                    Error(f'{lang.getstr("profile.invalid")}\n{path}'), self
                )
            else:
                setcfg("last_icc_path", path)
                setcfg("last_cal_or_icc_path", path)
        if profile:
            view_3d_format = getcfg("3d.format")
            if view_3d_format == "HTML":
                x3d = True
                html = True
            elif view_3d_format == "X3D":
                x3d = True
                html = False
            else:
                x3d = False
                html = False
            profile_path = None
            if profile.fileName and os.path.isfile(profile.fileName):
                profile_path = profile.fileName
            if not profile_path or not waccess(os.path.dirname(profile_path), os.W_OK):
                result = self.worker.create_tempdir()
                if isinstance(result, Exception):
                    show_result_dialog(result, self)
                    return
                desc = profile.getDescription()
                profile_path = os.path.join(
                    self.worker.tempdir,
                    f"{make_argyll_compatible_path(desc) }{profile_ext}",
                )
                profile.write(profile_path)
            profile_mtime = os.stat(profile_path).st_mtime
            filename, ext = os.path.splitext(profile_path)
            comparison_profile = self.gamut_view_options.comparison_profile
            comparison_profile_path = None
            comparison_profile_mtime = 0
            if comparison_profile:
                comparison_profile_path = comparison_profile.fileName
                if comparison_profile_path and not waccess(
                    os.path.dirname(comparison_profile_path), os.W_OK
                ):
                    result = self.worker.create_tempdir()
                    if isinstance(result, Exception):
                        show_result_dialog(result, self)
                        return
                    comparison_profile_path = os.path.join(
                        self.worker.tempdir,
                        make_argyll_compatible_path(
                            os.path.basename(comparison_profile_path)
                        ),
                    )
                    comparison_profile.write(comparison_profile_path)
                comparison_profile_mtime = os.stat(comparison_profile_path).st_mtime
            mods = []
            intent = self.gamut_view_options.intent
            if intent != "r":
                mods.append(intent)
            direction = self.gamut_view_options.direction[-1]
            if direction != "f":
                mods.append(direction)
            order = self.gamut_view_options.order
            if order != "n":
                mods.append(order)
            if mods:
                filename = "{} {}".format(
                    filename, "".join(["[{}]".format(mod.upper() for mod in mods)])
                )
            if comparison_profile_path:
                filename += (
                    " vs "
                    + os.path.splitext(os.path.basename(comparison_profile_path))[0]
                )
                if mods:
                    filename = "{} {}".format(
                        filename, "".join([f"[{mod.upper()}]" for mod in mods])
                    )
            for vrmlext in (".vrml", ".vrml.gz", ".wrl", ".wrl.gz", ".wrz"):
                vrmlpath = filename + vrmlext
                if sys.platform == "win32":
                    vrmlpath = make_win32_compatible_long_path(vrmlpath)
                if os.path.isfile(vrmlpath):
                    break
            outfilename = filename
            colorspace = self.gamut_view_options.get_colorspace(3)
            if colorspace != "Lab":
                outfilename = f"{outfilename} {colorspace}"
            vrmloutpath = f"{outfilename}{vrmlext}"
            x3dpath = f"{outfilename}.x3d"
            if html:
                finalpath = f"{x3dpath}.html"
            elif x3d:
                finalpath = x3dpath
            else:
                finalpath = vrmloutpath
            if sys.platform == "win32":
                vrmloutpath = make_win32_compatible_long_path(vrmloutpath)
                x3dpath = make_win32_compatible_long_path(x3dpath)
                finalpath = make_win32_compatible_long_path(finalpath)
            for outpath in (finalpath, vrmloutpath, vrmlpath):
                if os.path.isfile(outpath):
                    mtime = os.stat(outpath).st_mtime
                    if profile_mtime > mtime or comparison_profile_mtime > mtime:
                        # Profile(s) have changed, delete existing 3D visualization
                        # so it will be re-generated
                        os.remove(outpath)
            if os.path.isfile(finalpath):
                launch_file(finalpath)
            else:
                if os.path.isfile(vrmloutpath):
                    self.view_3d_consumer(
                        True, None, None, vrmloutpath, x3d, x3dpath, html
                    )
                elif os.path.isfile(vrmlpath):
                    self.view_3d_consumer(
                        True, colorspace, None, vrmlpath, x3d, x3dpath, html
                    )
                else:
                    # Create VRML file
                    profile_paths = [profile_path]
                    if comparison_profile_path:
                        profile_paths.append(comparison_profile_path)
                    self.worker.start(
                        self.view_3d_consumer,
                        self.worker.calculate_gamut,
                        cargs=(colorspace, filename, None, x3d, x3dpath, html),
                        wargs=(profile_paths, intent, direction, order, False),
                        progress_msg=lang.getstr("gamut.view.create"),
                        continue_next=x3d,
                        fancy=False,
                    )

    def view_3d_consumer(
        self, result, colorspace, filename, vrmlpath, x3d, x3dpath, html
    ):
        if filename:
            if getcfg("vrml.compress"):
                vrmlpath = f"{filename}.wrz"
            else:
                vrmlpath = f"{filename}.wrl"
        if os.path.isfile(vrmlpath) and colorspace not in ("Lab", None):
            filename, ext = os.path.splitext(vrmlpath)
            if ext.lower() in (".gz", ".wrz"):
                cls = GzipFileProper
            else:
                cls = open
            with cls(vrmlpath, "rb") as vrmlfile:
                vrml = vrmlfile.read()
            vrml = x3dom.update_vrml(vrml, colorspace)
            vrmlpath = f"{filename} {colorspace}{ext}"
            with cls(vrmlpath, "wb") as vrmlfile:
                vrmlfile.write(vrml)
        if not os.path.isfile(vrmlpath):
            if self.worker.errors:
                result = UnloggedError("".join(self.worker.errors).strip())
            else:
                result = Error(lang.getstr("file.missing", vrmlpath))
        if isinstance(result, Exception):
            self.worker.stop_progress()
            show_result_dialog(result, self)
        elif x3d:
            vrmlfile2x3dfile(
                vrmlpath,
                x3dpath,
                embed=getcfg("x3dom.embed"),
                html=html,
                view=True,
                force=False,
                cache=getcfg("x3dom.cache"),
                worker=self.worker,
            )
        else:
            launch_file(vrmlpath)

    def view_3d_format_popup(self, event):
        menu = wx.Menu()

        item_selected = False
        for file_format in config.valid_values["3d.format"]:
            item = menu.AppendRadioItem(-1, file_format)
            item.Check(file_format == getcfg("3d.format"))
            self.Bind(wx.EVT_MENU, self.view_3d_format_handler, id=item.Id)

        self.PopupMenu(menu)
        for item in menu.MenuItems:
            self.Unbind(wx.EVT_MENU, id=item.Id)
        menu.Destroy()

    def view_3d_format_handler(self, event):
        for item in event.EventObject.MenuItems:
            if item.IsChecked():
                setcfg("3d.format", item.GetItemLabelText())

        self.view_3d(None)


def main():
    config.initcfg("profile-info")
    lang.init()
    lang.update_defaults()
    app = BaseApp(0)
    app.TopWindow = ProfileInfoFrame(None, -1)
    if sys.platform == "darwin":
        app.TopWindow.init_menubar()
    wx.CallLater(1, _main, app)
    app.MainLoop()


def _main(app):
    app.TopWindow.listen()
    display_no = get_argyll_display_number(app.TopWindow.get_display()[1])
    for arg in sys.argv[1:]:
        if os.path.isfile(arg):
            app.TopWindow.drop_handler(str(os.path.abspath(arg)))
            break
    else:
        app.TopWindow.LoadProfile(get_display_profile(display_no))
    app.TopWindow.Show()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import codecs
import locale
import re
import string
import sys
import unicodedata
from functools import reduce

env_errors = (EnvironmentError,)
if sys.platform == "win32":
    import pywintypes

    env_errors = env_errors + (pywintypes.error, pywintypes.com_error)

from DisplayCAL.encoding import get_encodings

fs_enc = get_encodings()[1]

ascii_printable = "".join(
    [
        getattr(string, name)
        for name in ("digits", "ascii_letters", "punctuation", "whitespace")
    ]
)

# Control chars are defined as charcodes in the decimal range 0-31 (inclusive)
# except whitespace characters, plus charcode 127 (DEL)
control_chars = "".join(
    [chr(i) for i in list(range(0, 9)) + list(range(14, 32)) + [127]]
)

# Safe character substitution - can be used for filenames
# i.e. no \/:*?"<>| will be added through substitution
# Contains only chars that are not normalizable
safesubst = {
    # Latin-1 supplement
    "\u00a2": "c",  # Cent sign
    "\u00a3": "GBP",  # Pound sign
    "\u00a5": "JPY",  # Yen sign
    "\u00a9": "(C)",  # U+00A9 copyright sign
    "\u00ac": "!",  # Not sign
    "\u00ae": "(R)",  # U+00AE registered sign
    "\u00b0": "deg",  # Degree symbol
    "\u00b1": "+-",
    "\u00c4": "Ae",  # Capital letter A with diaresis (Umlaut)
    "\u00c5": "Aa",  # Capital letter A with ring above
    "\u00c6": "AE",
    "\u00d6": "Oe",  # Capital letter O with diaresis (Umlaut)
    "\u00dc": "Ue",  # Capital letter U with diaresis (Umlaut)
    "\u00d7": "x",  # U+00D7 multiplication sign
    "\u00df": "ss",
    "\u00e4": "ae",  # Small letter a with diaresis (Umlaut)
    "\u00e5": "aa",  # Small letter a with ring above
    "\u00e6": "ae",
    "\u00f6": "oe",  # Small letter o with diaresis (Umlaut)
    "\u00fc": "ue",  # Small letter u with diaresis (Umlaut)
    # Latin extended A
    "\u0152": "OE",
    "\u0153": "oe",
    # General punctuation
    "\u2010": "-",
    "\u2011": "-",
    "\u2012": "-",
    "\u2013": "-",  # U+2013 en dash
    "\u2014": "--",  # U+2014 em dash
    "\u2015": "---",  # U+2015 horizontal bar
    "\u2018": "'",
    "\u2019": "'",
    "\u201a": ",",
    "\u201b": "'",
    "\u201c": "''",
    "\u201d": "''",
    "\u201e": ",,",
    "\u201f": "''",
    "\u2032": "'",
    "\u2033": "''",
    "\u2034": "'''",
    "\u2035": "'",
    "\u2036": "''",
    "\u2037": "'''",
    "\u2053": "~",
    # Superscripts and subscripts
    "\u207b": "-",  # Superscript minus
    "\u208b": "-",  # Subscript minus
    # Currency symbols
    "\u20a1": "CRC",  # Costa Rica 'Colon'
    "\u20a6": "NGN",  # Nigeria 'Naira'
    "\u20a9": "KRW",  # South Korea 'Won'
    "\u20aa": "ILS",  # Isreael 'Sheqel'
    "\u20ab": "VND",  # Vietnam 'Dong'
    "\u20ac": "EUR",
    "\u20ad": "LAK",  # Laos 'Kip'
    "\u20ae": "MNT",  # Mongolia 'Tugrik'
    "\u20b2": "PYG",  # Paraguay 'Guarani'
    "\u20b4": "UAH",  # Ukraine 'Hryvnja'
    "\u20b5": "GHS",  # Ghana 'Cedi'
    "\u20b8": "KZT",  # Kasachstan 'Tenge'
    "\u20b9": "INR",  # Indian 'Rupee'
    "\u20ba": "TRY",  # Turkey 'Lira'
    "\u20bc": "AZN",  # Aserbaidchan 'Manat'
    "\u20bd": "RUB",  # Russia 'Ruble'
    "\u20be": "GEL",  # Georgia 'Lari'
    # Letter-like symbols
    "\u2117": "(P)",
    # Mathematical operators
    "\u2212": "-",  # U+2212 minus sign
    "\u2260": "!=",
    # Enclosed alphanumerics
    "\u2460": "(1)",
    "\u2461": "(2)",
    "\u2462": "(3)",
    "\u2463": "(4)",
    "\u2464": "(5)",
    "\u2465": "(6)",
    "\u2466": "(7)",
    "\u2467": "(8)",
    "\u2468": "(9)",
    "\u2469": "(10)",
    "\u246a": "(11)",
    "\u246b": "(12)",
    "\u246c": "(13)",
    "\u246d": "(14)",
    "\u246e": "(15)",
    "\u246f": "(16)",
    "\u2470": "(17)",
    "\u2471": "(18)",
    "\u2472": "(19)",
    "\u2473": "(20)",
    "\u24eb": "(11)",
    "\u24ec": "(12)",
    "\u24ed": "(13)",
    "\u24ee": "(14)",
    "\u24ef": "(15)",
    "\u24f0": "(16)",
    "\u24f1": "(17)",
    "\u24f2": "(18)",
    "\u24f3": "(19)",
    "\u24f4": "(20)",
    "\u24f5": "(1)",
    "\u24f6": "(2)",
    "\u24f7": "(3)",
    "\u24f8": "(4)",
    "\u24f9": "(5)",
    "\u24fa": "(6)",
    "\u24fb": "(7)",
    "\u24fc": "(8)",
    "\u24fd": "(9)",
    "\u24fe": "(10)",
    "\u24ff": "(0)",
    # Dingbats
    "\u2776": "(1)",
    "\u2777": "(2)",
    "\u2778": "(3)",
    "\u2779": "(4)",
    "\u277a": "(5)",
    "\u277b": "(6)",
    "\u277c": "(7)",
    "\u277d": "(8)",
    "\u277e": "(9)",
    "\u277f": "(10)",
    "\u2780": "(1)",
    "\u2781": "(2)",
    "\u2782": "(3)",
    "\u2783": "(4)",
    "\u2784": "(5)",
    "\u2785": "(6)",
    "\u2786": "(7)",
    "\u2787": "(8)",
    "\u2788": "(9)",
    "\u2789": "(10)",
    "\u278a": "(1)",
    "\u278b": "(2)",
    "\u278c": "(3)",
    "\u278d": "(4)",
    "\u278e": "(5)",
    "\u278f": "(6)",
    "\u2790": "(7)",
    "\u2791": "(8)",
    "\u2792": "(9)",
    "\u2793": "(10)",
}

# Extended character substitution - can NOT be used for filenames
# Contains only chars that are not normalizable
subst = dict(safesubst)
subst.update(
    {
        # Latin-1 supplement
        "\u00a6": "|",
        "\u00ab": "<<",
        "\u00bb": ">>",
        "\u00bc": "1/4",
        "\u00bd": "1/2",
        "\u00be": "3/4",
        "\u00f7": ":",
        # General punctuation
        "\u201c": '"',
        "\u201d": '"',
        "\u201f": '"',
        "\u2033": '"',
        "\u2036": '"',
        "\u2039": "<",
        "\u203a": ">",
        "\u203d": "!?",
        "\u2044": "/",
        # Number forms
        "\u2153": "1/3",
        "\u2154": "2/3",
        "\u215b": "1/8",
        "\u215c": "3/8",
        "\u215d": "5/8",
        "\u215e": "7/8",
        # Arrows
        "\u2190": "<-",
        "\u2192": "->",
        "\u2194": "<->",
        # Mathematical operators
        "\u226a": "<<",
        "\u226b": ">>",
        "\u2264": "<=",
        "\u2265": "=>",
    }
)


class StrList(list):

    """It's a list. It's a string. It's a list of strings that behaves like a
    string! And like a list."""

    def __init__(self, seq=tuple()):
        list.__init__(self, seq)

    def __iadd__(self, text):
        self.append(text)
        return self

    def __getattr__(self, attr):
        return getattr(str(self), attr)

    def __str__(self):
        return "".join(self)


def asciize(obj):
    """Turn several unicode chars into an ASCII representation.

    This function either takes a string or an exception as argument (when used
    as error handler for encode or decode).
    """
    chars = b""
    if isinstance(obj, Exception):
        for char in obj.object[obj.start: obj.end]:
            chars += subst.get(char, normalencode(char).strip() or b"?")
        return chars, obj.end
    else:
        return obj.encode("ASCII", "asciize")


codecs.register_error("asciize", asciize)


def safe_asciize(obj):
    """Turn several unicode chars into an ASCII representation.

    This function either takes a string or an exception as argument (when used
    as error handler for encode or decode).
    """
    chars = b""
    if isinstance(obj, Exception):
        for char in obj.object[obj.start: obj.end]:
            if char in safesubst:
                subst_char = safesubst[char]
            else:
                subst_char = b"_"
                if char not in subst:
                    subst_char = normalencode(char).strip() or subst_char
            chars += subst_char
        return chars, obj.end
    elif isinstance(obj, str):
        return obj.encode("ASCII", "safe_asciize")
    elif isinstance(obj, bytes):
        return obj.decode("utf-8").encode("ASCII", "safe_asciize")


codecs.register_error("safe_asciize", safe_asciize)


def escape(obj):
    """Turn unicode chars into escape codes.

    This function either takes a string or an exception as argument (when used
    as error handler for encode or decode).

    """
    chars = b""
    if isinstance(obj, Exception):
        for char in obj.object[obj.start: obj.end]:
            chars += subst.get(char, "\\u%s" % hex(ord(char))[2:].rjust(4, "0"))
        return chars, obj.end
    else:
        return obj.encode("ASCII", "escape")


# TODO: convert this to a decorator
codecs.register_error("escape", escape)


def make_ascii_printable(text, substitute=b""):
    """Make ASCII printable.

    Args:
        text (bytes, str): A ``bytes`` or ``str`` value.
        substitute (bytes, str): A ``bytes`` or ``str`` value.

    Returns:
        bytes or str:
    """
    buffer = []
    joiner = ""
    temp_ascii_printable = ascii_printable
    if isinstance(text, bytes):
        joiner = b""
        temp_ascii_printable = ascii_printable.encode("utf-8")
        if isinstance(substitute, str):
            substitute = substitute.encode("utf-8")
    else:
        if isinstance(substitute, bytes):
            substitute = substitute.decode("utf-8")

    for i in range(len(text)):
        char = text[i: i + 1]
        if char in temp_ascii_printable:
            buffer.append(char)
        else:
            buffer.append(substitute)
    return joiner.join(buffer)


def make_filename_safe(unistr, encoding=fs_enc, substitute="_", concat=True):
    """Make sure string is safe to use as filename.

    I.e. turn characters that are invalid in the filesystem encoding into ASCII
    equivalents and replace characters that are invalid in filenames with
    substitution character.
    """
    if not isinstance(unistr, (str, bytes)):
        raise TypeError("unistr should be a str or bytes, not {}".format(
            unistr.__class__.__name__
        ))
    # Turn characters that are invalid in the filesystem encoding into ASCII
    # substitution character '?'
    # NOTE that under Windows, encoding with the filesystem encoding may
    # substitute some characters even in "strict" replacement mode depending
    # on the Windows language setting for non-Unicode programs! (Python 2.x
    # under Windows supports Unicode by wrapping the win32 ASCII API, so it is
    # a non-Unicode program from that point of view. This problem presumably
    # doesn't exist with Python 3.x which uses the win32 Unicode API)
    if isinstance(unistr, str):
        # str
        unidec = unistr.encode(encoding, "replace").decode(encoding, "replace")
        pattern = r"[\\/:*?\"<>|]"
        plus = "+"
        uniout = ""
        question_mark = "?"
        if isinstance(substitute, bytes):
            substitute = substitute.decode(encoding)
    else:
        # bytes
        unidec = unistr.decode(encoding, "replace").encode(encoding, "replace")
        pattern = rb"[\\/:*?\"<>|]"
        plus = b"+"
        uniout = b""
        question_mark = b"?"
        if isinstance(substitute, str):
            substitute = substitute.encode(encoding)

    # Replace substitution character '?' with ASCII equivalent of original char
    for i in range(len(unidec)):
        c = unidec[i: i + 1]
        if c == question_mark:
            c = safe_asciize(unistr[i: i + 1])  # this will always be bytes
            if isinstance(unistr, str):
                c = c.decode(encoding)
        uniout += c
    # Remove invalid chars
    if concat:
        pattern += plus

    uniout = re.sub(pattern, substitute, uniout)
    return uniout


def normalencode(unistr, form="NFKD", encoding="ASCII", errors="ignore"):
    """Return encoded normal form of unicode string"""
    return unicodedata.normalize(form, unistr).encode(encoding, errors)


def box(text, width=80, collapse=False):
    """Create a box around text (monospaced font required for display)"""
    content_width = width - 4
    text = wrap(str(text), content_width)
    lines = text.splitlines()
    if collapse:
        content_width = 0
        for line in lines:
            content_width = max(len(line), content_width)
        width = content_width + 4
    horizontal_line = "\u2500" * (width - 2)
    box = ["\u250c%s\u2510" % horizontal_line]
    for line in lines:
        box.append("\u2502 %s \u2502" % line.ljust(content_width))
    box.append("\u2514%s\u2518" % horizontal_line)
    return "\n".join(box)


def center(text, width=None):
    """Center (mono-spaced) text.

    If no width is given, the longest line
    (after breaking at each newline) is used as max width.

    """
    text = text.split("\n")
    if width is None:
        width = 0
        for line in text:
            if len(line) > width:
                width = len(line)
    i = 0
    for line in text:
        text[i] = line.center(width)
        i += 1
    return "\n".join(text)


def create_replace_function(template, values):
    """Create a replace function for use with e.g. re.sub"""

    def replace_function(match, template=template, values=values):
        template = match.expand(template)
        return template % values

    return replace_function


def ellipsis_(text, maxlen=64, pos="r"):
    """Truncate text to maxlen characters and add ellipsis if it was longer.

    Ellipsis position can be 'm' (middle) or 'r' (right).
    """
    if len(text) <= maxlen:
        return text
    ellipsis_char = "\u2026"
    if isinstance(text, bytes):
        ellipsis_char = b"u\xc2\x826"
    if pos == "r":
        return text[: maxlen - 1] + ellipsis_char
    elif pos == "m":
        return text[: int(maxlen / 2)] + ellipsis_char + text[int(-maxlen / 2 + 1):]


def hexunescape(match):
    """To be used with re.sub"""
    return chr(int(match.group(1), 16))


def indent(text, prefix, predicate=None):
    # From Python 3.7 textwrap module
    """Adds 'prefix' to the beginning of selected lines in 'text'.

    If 'predicate' is provided, 'prefix' will only be added to the lines
    where 'predicate(line)' is True. If 'predicate' is not provided,
    it will default to adding 'prefix' to all non-empty lines that do not
    consist solely of whitespace characters.
    """
    if predicate is None:

        def predicate(line):
            return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)

    return "".join(prefixed_lines())


def universal_newlines(txt):
    """Return txt with all new line formats converted to POSIX newlines."""
    if isinstance(txt, str):
        return txt.replace("\r\n", "\n").replace("\r", "\n")
    elif isinstance(txt, bytes):
        return txt.replace(b"\r\n", b"\n").replace(b"\r", b"\n")


def replace_control_chars(txt, replacement=" ", collapse=False):
    """Replace all control characters.

    Default replacement character is ' ' (space).
    If the 'collapse' keyword argument evaluates to True, consecutive
    replacement characters are collapsed to a single one.

    """
    txt = strtr(txt, dict(list(zip(control_chars, [replacement] * len(control_chars)))))
    if collapse:
        while replacement * 2 in txt:
            txt = txt.replace(replacement * 2, replacement)
    return txt


def safe_basestring(obj, enc="utf-8", errors="replace"):
    """Return a unicode or string representation of obj

    Return obj if isinstance(obj, basestring). Otherwise, return unicode(obj),
    string(obj), or repr(obj), whichever succeeds first.
    """
    if isinstance(obj, env_errors):
        # Possible variations of environment-type errors:
        # - instance with 'errno', 'strerror', 'filename' and 'args' attributes
        #   (created by EnvironmentError with three arguments)
        #   NOTE: The 'args' attribute will contain only the first two arguments
        # - instance with 'errno', 'strerror' and 'args' attributes
        #   (created by EnvironmentError with two arguments)
        # - instance with just 'args' attribute
        #   (created by EnvironmentError with one or more than three arguments)
        # - urllib2.URLError with empty 'args' attribute but 'reason' and
        #   'filename' attributes
        # - pywintypes.error with 'funcname', 'message', 'strerror', 'winerror'
        #   and 'args' attributes
        if hasattr(obj, "reason"):
            if isinstance(obj.reason, str):
                obj.args = (obj.reason,)
            elif isinstance(obj.reason, bytes):
                obj.args = (str(obj.reason, enc, errors),)
            else:
                obj.args = obj.reason
        error = []
        if getattr(obj, "winerror", None) is not None:
            # pywintypes.error or WindowsError
            error.append("[Windows Error %s]" % obj.winerror)
        elif getattr(obj, "errno", None) is not None:
            error.append("[Errno %s]" % obj.errno)
        if getattr(obj, "strerror", None) is not None:
            if getattr(obj, "filename", None) is not None:
                error.append(obj.strerror.rstrip(":.") + ":")
            elif getattr(obj, "funcname", None) is not None:
                # pywintypes.error
                error.append(obj.funcname + ": " + obj.strerror)
            else:
                error.append(obj.strerror)
        if not error:
            error = list(obj.args)
        if getattr(obj, "filename", None) is not None:
            error.append(obj.filename)

        temp_error = []
        for arg in error:
            if isinstance(arg, str):
                temp_error.append(arg)
            elif isinstance(arg, bytes):
                temp_error.append(str(arg, enc, errors))
            else:
                temp_error.append(str(arg))
        error = temp_error
        obj = " ".join(error)
    elif isinstance(obj, KeyError) and obj.args:
        obj = "Key does not exist: " + repr(obj.args[0])
    oobj = obj
    if not isinstance(obj, str):
        try:
            obj = str(obj, enc, errors)
        except (UnicodeDecodeError, TypeError):
            obj = repr(obj)

    if isinstance(oobj, Exception) and not isinstance(oobj, Warning):
        module = getattr(oobj, "__module__", "")
        package = safe_basestring.__module__.split(".")[0]  # Our own package
        if not module.startswith(package + "."):
            clspth = ".".join([_f for _f in [module, oobj.__class__.__name__] if _f])
            if not obj.startswith(clspth + ":") and obj != clspth:
                obj = ": ".join([_f for _f in [clspth, obj] if _f])
    return obj


def safe_str(obj, enc=fs_enc, errors="replace"):
    """Return string representation of obj"""
    return safe_basestring(obj, enc=enc, errors=errors)


def strtr(txt, replacements):
    """String multi-replace, a bit like PHP's strtr.

    replacements can be a dict, a list or a string.
    If list or string, all matches are replaced with the empty string ("").

    """
    if hasattr(replacements, "items"):
        replacements = iter(replacements.items())
    elif isinstance(replacements, str):
        for srch in replacements:
            txt = txt.replace(srch, "")
    elif isinstance(replacements, bytes):
        for srch in replacements:
            txt = txt.replace(srch, b"")
        return txt
    for srch, sub in replacements:
        if isinstance(txt, bytes):
            if not isinstance(srch, bytes):
                srch = srch.encode("utf-8")
            if not isinstance(sub, bytes):
                sub = sub.encode("utf-8")
        elif isinstance(txt, str):
            if not isinstance(srch, str):
                srch = srch.decode("utf-8")
            if not isinstance(sub, str):
                sub = sub.decode("utf-8")
        txt = txt.replace(srch, sub)
    return txt


def wrap(text, width=70):
    """A word-wrap function that preserves existing line breaks and spaces.

    Expects that existing line breaks are posix newlines (\\n).

    """
    return reduce(
        lambda line, word, width=width: "%s%s%s"
        % (
            line,
            " \n"[
                (
                    len(line) - line.rfind("\n") - 1 + len(word.split("\n", 1)[0])
                    >= width
                )
            ],
            word,
        ),
        text.split(" "),
    )


def test():
    for k in subst:
        v = subst[k]
        print(k, v)


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
"""
Meta information
"""

import re
import sys


try:
    from DisplayCAL.__version__ import (
        BUILD_DATE as build,
        LASTMOD as lastmod,
        VERSION,
        VERSION_BASE,
        VERSION_STRING,
    )
except ImportError:
    build = lastmod = "1970-01-01T00:00:00Z"
    VERSION = None
    VERSION_STRING = None

from DisplayCAL.options import test_update

if not VERSION or test_update:
    VERSION = VERSION_BASE = (0, 0, 0)
    VERSION_STRING = ".".join(str(n) for n in VERSION)

author = ", ".join([
    "Florian Höch",
    "Erkan Özgür Yılmaz",
    "Patrick Zwerschke"
])
author_ascii = ", ".join([
    "Florian Hoech",
    "Erkan Ozgur Yilmaz",
    "Patrick Zwerschke"
])
description = (
    "Display calibration and profiling with a focus on accuracy and versatility"
)
longdesc = (
    "Calibrate and characterize your display devices using one of many supported "
    "measurement instruments, with support for multi-display setups and a variety of "
    "available options for advanced users, such as  verification and reporting "
    "functionality to evaluate ICC profiles and display devices, creating video 3D "
    "LUTs, as well as optional CIECAM02 gamut mapping to take into account varying "
    "viewing conditions."
)
DOMAIN = "displaycal.net"
development_home_page = "https://github.com/eoyilmaz/displaycal-py3"

author_email = ", ".join(
    [
        f"florian{chr(0o100)}{DOMAIN}",
        f"eoyilmaz{chr(0o100)}gmail.com",
        f"patrick{chr(0o100)}p5k.org",
    ]
)
name = "DisplayCAL"
appstream_id = ".".join(reversed([name] + DOMAIN.split(".")))
name_html = '<span class="appname">Display<span>CAL</span></span>'

py_minversion = (3, 8)
py_maxversion = (3, 11)

version = VERSION_STRING
version_lin = VERSION_STRING  # Linux
version_mac = VERSION_STRING  # Mac OS X
version_win = VERSION_STRING  # Windows
version_src = VERSION_STRING
version_short = re.sub(r"(?:\.0){1,2}$", "", version)
version_tuple = VERSION  # only ints allowed and must be exactly 3 values

wx_minversion = (2, 8, 11)
wx_recversion = (4, 1, 1)


def get_latest_changelog_entry(readme):
    """Get changelog entry for latest version from ReadMe HTML"""
    changelog = re.search(
        r'<div id="(?:changelog|history)">.+?<h2>.+?</h2>.+?<dl>.+?</dd>', readme, re.S
    )

    if changelog:
        changelog = changelog.group()
        changelog = re.sub(r'\s*<div id="(?:changelog|history)">\n?', "", changelog)
        changelog = re.sub(r"\s*</?d[ld]>\n?", "", changelog)
        changelog = re.sub(r"\s*<(h[23])>.+?</\1>\n?", "", changelog)

    return changelog


def script2pywname(script):
    """Convert all-lowercase script name to mixed-case pyw name"""
    a2b = {
        name + "-3dlut-maker": name + "-3DLUT-maker",
        name + "-vrml-to-x3d-converter": name + "-VRML-to-X3D-converter",
        name + "-eecolor-to-madvr-converter": name + "-eeColor-to-madVR-converter",
    }
    if script.lower().startswith(name.lower()):
        pyw = name + script[len(name) :]
        return a2b.get(pyw, pyw)
    return script



# -*- coding: utf-8 -*-

from time import sleep

# 0install: Make sure imported protobuf is from implementation to ensure
# correct version
import sys

if not getattr(sys, "frozen", False):
    import os
    import re

    for pth in sys.path:
        if os.path.basename(pth).startswith("protobuf-") and re.match(
            r"sha\d+(?:new)?", os.path.basename(os.path.dirname(pth))
        ):
            if "google" in sys.modules:
                del sys.modules["google"]
            try:
                import pkg_resources
            except ImportError:
                import pkgutil
            syspath = sys.path[:]
            sys.path[:] = [pth]
            import google.protobuf

            sys.path[:] = syspath
            break

if sys.version_info[:2] < (3,):
    # zeroconf 0.19.1 is the last version supporting Python 2.x
    # but zeroconf 0.20 lacks a check for Python version
    import zeroconf

    if zeroconf.__version__ > "0.19.1":
        raise ImportError(
            """
Python version > 3.3 required for python-zeroconf %s.
If you need support for Python 2 or Python 3.3 please use version 0.19.1
    """
            % zeroconf.__version__
        )

from pychromecast import get_chromecasts
from pychromecast.controllers import BaseController

from DisplayCAL import localization as lang


class ChromeCastPatternGeneratorController(BaseController):
    def __init__(self):
        super(ChromeCastPatternGeneratorController, self).__init__(
            "urn:x-cast:net.hoech.cast.patterngenerator", "B5C2CBFC"
        )
        self.request_id = 0

    def receive_message(self, message, data):
        return True  # Indicate we handled this message

    def send(
        self,
        rgb=(0, 0, 0),
        bgrgb=(0, 0, 0),
        offset_x=0.5,
        offset_y=0.5,
        h_scale=1,
        v_scale=1,
    ):
        fg = "#%02X%02X%02X" % tuple(round(v * 255) for v in rgb)
        bg = "#%02X%02X%02X" % tuple(round(v * 255) for v in bgrgb)
        self.request_id += 1
        self.send_message(
            {
                "requestId": self.request_id,
                "foreground": fg,
                "offset": [offset_x, offset_y],
                "scale": [h_scale, v_scale],
                "background": bg,
            }
        )


class ChromeCastPatternGenerator(object):
    def __init__(self, name, logfile=None):
        self._controller = ChromeCastPatternGeneratorController()
        self.name = name
        self.listening = False
        self.logfile = logfile

    def disconnect_client(self):
        self.listening = False
        if hasattr(self, "_cc"):
            if self._cc.app_id:
                self._cc.quit_app()
        if hasattr(self, "conn"):
            del self.conn

    def send(
        self,
        rgb=(0, 0, 0),
        bgrgb=(0, 0, 0),
        bits=None,
        use_video_levels=None,
        x=0,
        y=0,
        w=1,
        h=1,
    ):
        if w < 1:
            x /= 1.0 - w
        else:
            x = 0
        if h < 1:
            y /= 1.0 - h
        else:
            y = 0
        self._controller.send(rgb, bgrgb, x, y, w * 10, h * 10)

    def wait(self):
        self.listening = True
        if self.logfile:
            self.logfile.write(
                lang.getstr("connecting.to", ("Chromecast", " " + self.name)) + "\n"
            )
        if not hasattr(self, "_cc"):
            # Find our ChromeCast
            try:
                self._cc = next(
                    cc
                    for cc in get_chromecasts()
                    if str(cc.device.friendly_name) == self.name
                )
            except StopIteration:
                self.listening = False
                raise
            self._cc.register_handler(self._controller)
        # Wait for ChromeCast device to be ready
        while self.listening:
            self._cc.wait(0.05)
            if self._cc.status:
                break
        if self.listening:
            # Launch pattern generator app
            self._controller.launch()
            # Wait for it
            while (
                self.listening and self._cc.app_id != self._controller.supporting_app_id
            ):
                sleep(0.05)
            self.conn = True
            print(lang.getstr("connection.established"))


if __name__ == "__main__":
    from DisplayCAL import config

    config.initcfg()
    lang.init()

    pg = ChromeCastPatternGenerator("Smörebröd")

    # Find our ChromeCast and connect to it, then launch the pattern generator
    # app on the ChromeCast device
    pg.wait()

    # Display test color (yellow window on blue background)
    pg.send((1, 1, 0), (0, 0, 1), x=0.375, y=0.375, w=0.25, h=0.25)



# -*- coding: utf-8 -*-

import decimal
import math


def float2dec(f, digits=10):
    parts = str(f).split(".")
    if len(parts) > 1:
        if parts[1][:digits] == "9" * digits:
            f = math.ceil(f)
        elif parts[1][:digits] == "0" * digits:
            f = math.floor(f)
    return decimal.Decimal(str(f))


def stripzeros(n):
    """Strip zeros and convert to decimal.

    Will always return the shortest decimal representation
    (1.0 becomes 1, 1.234567890 becomes 1.23456789).

    """
    if isinstance(n, (float, int)):
        n = "%.10f" % n
    else:
        n = str(n)
    if "." in n:
        n = n.rstrip("0").rstrip(".")
    try:
        n = decimal.Decimal(n)
    except decimal.InvalidOperation:
        pass
    return n



# -*- coding: utf-8 -*-

import wx
import wx.xrc as xrc

try:
    from DisplayCAL import floatspin
except ImportError:
    import wx.lib.agw.floatspin as floatspin


class FloatSpinCtrlXmlHandler(xrc.XmlResourceHandler):
    def __init__(self):
        xrc.XmlResourceHandler.__init__(self)
        # Standard styles
        self.AddWindowStyles()

    def CanHandle(self, node):
        return self.IsOfClass(node, "FloatSpin")

    # Process XML parameters and create the object
    def DoCreateResource(self):
        try:
            min_val = float(self.GetText("min_val"))
        except Exception:
            min_val = None
        try:
            max_val = float(self.GetText("max_val"))
        except Exception:
            max_val = None
        try:
            increment = float(self.GetText("increment"))
        except Exception:
            increment = 1.0
        is_spinctrldbl = hasattr(wx, "SpinCtrlDouble") and issubclass(
            floatspin.FloatSpin, wx.SpinCtrlDouble
        )
        if is_spinctrldbl:
            defaultstyle = wx.SP_ARROW_KEYS | wx.ALIGN_RIGHT
        else:
            defaultstyle = 0
        w = floatspin.FloatSpin(
            parent=self.GetParentAsWindow(),
            id=self.GetID(),
            pos=self.GetPosition(),
            size=self.GetSize(),
            style=self.GetStyle(defaults=defaultstyle),
            min_val=min_val,
            max_val=max_val,
            increment=increment,
            name=self.GetName(),
        )

        try:
            w.SetValue(float(self.GetText("value")))
        except Exception:
            pass

        try:
            w.SetDigits(int(self.GetText("digits")))
        except Exception:
            pass

        self.SetupWindow(w)
        return w



# -*- coding: utf-8 -*-


from binascii import hexlify
import atexit
import math
import os
import pipes
import re
import shutil
import struct
import subprocess as sp
import sys
import tempfile
import textwrap
import traceback

if sys.platform == "win32":
    import win32api

from DisplayCAL.argyll_names import (
    names as argyll_names,
    altnames as argyll_altnames,
    optional as argyll_optional,
)
from DisplayCAL.colormath import (
    VidRGB_to_eeColor,
    VidRGB_to_cLUT65,
    cLUT65_to_VidRGB,
    eeColor_to_VidRGB,
)
from DisplayCAL.config import exe_ext, fs_enc, get_data_path, getcfg, profile_ext
from DisplayCAL.debughelpers import (
    Error,
    Info,
    UnloggedError,
    UnloggedInfo,
    UnloggedWarning,
    Warn,
)
from DisplayCAL.log import LogFile
from DisplayCAL.meta import name as appname
from DisplayCAL.multiprocess import mp, pool_slice
from DisplayCAL.options import debug, verbose
from DisplayCAL.util_os import getenvu, quote_args, which
from DisplayCAL.util_str import make_filename_safe, safe_basestring, safe_str
from DisplayCAL import CGATS
from DisplayCAL import colormath
from DisplayCAL import config
from DisplayCAL import ICCProfile as ICCP
from DisplayCAL import localization as lang


def _mp_xicclu(
    chunk,
    thread_abort_event,
    progress_queue,
    profile_filename,
    intent="r",
    direction="f",
    order="n",
    pcs=None,
    scale=1,
    cwd=None,
    startupinfo=None,
    use_icclu=False,
    use_cam_clipping=False,
    logfile=None,
    show_actual_if_clipped=False,
    input_encoding=None,
    output_encoding=None,
    abortmessage="Aborted",
    output_format=None,
    reverse=False,
    convert_video_rgb_to_clut65=False,
    verbose=1,
):
    if not config.cfg.items(config.configparser.DEFAULTSECT):
        config.initcfg()
    profile = ICCP.ICCProfile(profile_filename)
    xicclu = Xicclu(
        profile,
        intent,
        direction,
        order,
        pcs,
        scale,
        cwd,
        startupinfo,
        use_icclu,
        use_cam_clipping,
        logfile,
        None,
        show_actual_if_clipped,
        input_encoding,
        output_encoding,
        convert_video_rgb_to_clut65,
        verbose,
    )
    prevperc = 0
    start = 0
    num_subchunks = 50
    subchunksize = float(len(chunk)) / num_subchunks
    for i in range(num_subchunks):
        if (
            thread_abort_event is not None
            and getattr(sys, "_sigbreak", False)
            and not thread_abort_event.is_set()
        ):
            thread_abort_event.set()
            print("Got SIGBREAK, aborting thread...")
        if thread_abort_event is not None and thread_abort_event.is_set():
            xicclu.exit(raise_exception=False)
            return Info(abortmessage)
        end = int(math.ceil(subchunksize * (i + 1)))
        xicclu(chunk[start:end])
        start = end
        perc = round((i + 1.0) / num_subchunks * 100)
        if progress_queue and perc > prevperc:
            progress_queue.put(perc - prevperc)
            prevperc = perc
    xicclu.exit()
    return xicclu.get(output_format=output_format, reverse=reverse)


def _mp_generate_B2A_clut(
    chunk,
    thread_abort_event,
    progress_queue,
    profile_filename,
    intent,
    direction,
    pcs,
    use_cam_clipping,
    clutres,
    step,
    threshold,
    threshold2,
    interp,
    Linterp,
    m2,
    XYZbp,
    XYZwp,
    bpc,
    abortmessage="Aborted",
):
    """B2A cLUT generation worker

    This should be spawned as a multiprocessing process

    """
    if debug:
        print("comtypes?", "comtypes" in str(list(sys.modules.keys())))
        print("numpy?", "numpy" in str(list(sys.modules.keys())))
        print("wx?", "wx" in str(list(sys.modules.keys())))
        print("x3dom?", "x3dom" in str(list(sys.modules.keys())))
    if not config.cfg.items(config.configparser.DEFAULTSECT):
        config.initcfg()
    idata = []
    abmaxval = 255 + (255 / 256.0)
    profile = ICCP.ICCProfile(profile_filename)
    xicclu1 = Xicclu(profile, intent, direction, "n", pcs, 100)
    xicclu2 = xicclu1
    if use_cam_clipping:
        # Use CAM Jab for clipping for cLUT grid points after a given
        # threshold
        xicclu2 = Xicclu(
            profile, intent, direction, "n", pcs, 100, use_cam_clipping=True
        )
    prevperc = 0
    count = 0
    chunksize = len(chunk)
    for interp_tuple in (interp, Linterp):
        if interp_tuple:
            # Use numpy for speed
            if interp_tuple is interp:
                interp_list = list(interp_tuple)
            else:
                interp_list = [interp_tuple]
            for i, ointerp in enumerate(interp_list):
                interp_list[i] = colormath.Interp(
                    ointerp.xp, ointerp.fp, use_numpy=True
                )
            if interp_tuple is interp:
                interp = interp_list
            else:
                Linterp = interp_list[0]
    m2i = m2
    if profile.connectionColorSpace == b"XYZ":
        m2i = m2.inverted()
    for a in chunk:
        if thread_abort_event.is_set():
            if use_cam_clipping:
                xicclu2.exit()
            xicclu1.exit()
            return Info(abortmessage)
        for b in range(clutres):
            for c in range(clutres):
                d, e, f = [v * step for v in (a, b, c)]
                if profile.connectionColorSpace == b"XYZ":
                    # Apply TRC to XYZ values to distribute them optimally
                    # across cLUT grid points.
                    XYZ = [interp[i](v) for i, v in enumerate((d, e, f))]
                    # print "%3.6f %3.6f %3.6f" % tuple(XYZ), '->',
                    # Scale into PCS
                    v = m2i * XYZ
                    if bpc and XYZbp != [0, 0, 0]:
                        v = colormath.blend_blackpoint(v[0], v[1], v[2], None, XYZbp)
                    # print "%3.6f %3.6f %3.6f" % tuple(v)
                    # raw_input()
                    if intent == "a":
                        v = colormath.adapt(
                            *v + [XYZwp, list(profile.tags.wtpt.ir.values())]
                        )
                else:
                    # Legacy CIELAB
                    L = Linterp(d * 100)
                    v = L, -128 + e * abmaxval, -128 + f * abmaxval
                idata.append("%.6f %.6f %.6f" % tuple(v))
                # Lookup CIE -> device values through profile using xicclu
                if not use_cam_clipping or (
                    pcs == "x" and a <= threshold and b <= threshold and c <= threshold
                ):
                    xicclu1(v)
                if use_cam_clipping and (
                    pcs == "l" or a > threshold2 or b > threshold2 or c > threshold2
                ):
                    xicclu2(v)
                count += 1.0
            perc = round(count / (chunksize * clutres**2) * 100)
            if progress_queue and perc > prevperc:
                progress_queue.put(perc - prevperc)
                prevperc = perc
    if use_cam_clipping:
        xicclu2.exit()
        data2 = xicclu2.get()
    else:
        data2 = []
    xicclu1.exit()
    data1 = xicclu1.get()
    return idata, data1, data2


def check_argyll_bin(paths=None):
    """Check if the Argyll binaries can be found."""
    prev_dir = None
    cur_dir = os.curdir
    for name in argyll_names:
        exe = get_argyll_util(name, paths)
        if not exe:
            if name in argyll_optional:
                continue
            return False
        cur_dir = os.path.dirname(exe)
        if prev_dir:
            if cur_dir != prev_dir:
                if name in argyll_optional:
                    if verbose:
                        print(
                            "Warning: Optional Argyll executable %s is not in the same "
                            "directory as the main executables (%s)." % (exe, prev_dir)
                        )
                else:
                    if verbose:
                        print(
                            "Error: Main Argyll executable %s is not in the same "
                            "directory as the other executables (%s)." % (exe, prev_dir)
                        )
                    return False
        else:
            prev_dir = cur_dir
    if verbose >= 3:
        print("Argyll binary directory:", cur_dir)
    if debug:
        print("[D] check_argyll_bin OK")
    if debug >= 2:
        if not paths:
            paths = getenvu("PATH", os.defpath).split(os.pathsep)
            argyll_dir = (getcfg("argyll.dir") or "").rstrip(os.path.sep)
            if argyll_dir:
                if argyll_dir in paths:
                    paths.remove(argyll_dir)
                paths = [argyll_dir] + paths
        print("[D] Searchpath:\n  ", "\n  ".join(paths))
    # Fedora doesn't ship Rec709.icm
    config.defaults["3dlut.input.profile"] = (
        get_data_path(os.path.join("ref", "Rec709.icm"))
        or get_data_path(os.path.join("ref", "sRGB.icm"))
        or ""
    )
    config.defaults["testchart.reference"] = (
        get_data_path(os.path.join("ref", "ColorChecker.cie")) or ""
    )
    config.defaults["gamap_profile"] = (
        get_data_path(os.path.join("ref", "sRGB.icm")) or ""
    )
    return True


argyll_utils = {}


def get_argyll_util(name, paths=None):
    """Find a single Argyll utility. Return the full path."""
    cfg_argyll_dir = getcfg("argyll.dir")
    if not paths:
        paths = getenvu("PATH", os.defpath).split(os.pathsep)
        argyll_dir = (cfg_argyll_dir or "").rstrip(os.path.sep)
        if argyll_dir:
            if argyll_dir in paths:
                paths.remove(argyll_dir)
            paths = [argyll_dir] + paths
    cache_key = os.pathsep.join(paths)
    exe = argyll_utils.get(cache_key, {}).get(name, None)
    if exe:
        return exe
    elif verbose >= 4:
        print("Info: Searching for", name, "in", os.pathsep.join(paths))
    for path in paths:
        for altname in argyll_altnames.get(name, []):
            exe = which(altname + exe_ext, [path])
            if exe:
                break
        if exe:
            break
    if verbose >= 4:
        if exe:
            print("Info:", name, "=", exe)
        else:
            print(
                "Info:",
                "|".join(argyll_altnames[name]),
                "not found in",
                os.pathsep.join(paths),
            )
    if exe:
        if cache_key not in argyll_utils:
            argyll_utils[cache_key] = {}
        argyll_utils[cache_key][name] = exe
    return exe


def get_argyll_utilname(name, paths=None):
    """Find a single Argyll utility. Return the basename without extension."""
    exe = get_argyll_util(name, paths)
    if exe:
        exe = os.path.basename(os.path.splitext(exe)[0])
    return exe


def get_argyll_version(name, paths=None):
    """Determine version of a certain Argyll utility."""
    argyll_version_string = get_argyll_version_string(name, paths)
    return parse_argyll_version_string(argyll_version_string)


def get_argyll_version_string(name, paths=None):
    argyll_version_string = b"0.0.0"
    cmd = get_argyll_util(name, paths)
    if sys.platform == "win32":
        startupinfo = sp.STARTUPINFO()
        startupinfo.dwFlags |= sp.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = sp.SW_HIDE
    else:
        startupinfo = None
    try:
        p = sp.Popen(
            [cmd.encode(fs_enc), "-?"],
            stdin=sp.PIPE,
            stdout=sp.PIPE,
            stderr=sp.STDOUT,
            startupinfo=startupinfo,
        )
    except Exception as exception:
        print(exception)
        return argyll_version_string
    for line in (p.communicate(timeout=30)[0] or "").splitlines():
        if isinstance(line, bytes):
            line = line.strip()
            if b"version" in line.lower():
                argyll_version_string = line[line.lower().find(b"version") + 8:]
                break
    return argyll_version_string.decode("utf-8")


def parse_argyll_version_string(argyll_version_string):
    argyll_version = re.findall(r"(\d+|[^.\d]+)", argyll_version_string)
    for i, v in enumerate(argyll_version):
        try:
            argyll_version[i] = int(v)
        except ValueError:
            pass
    return argyll_version


def printcmdline(cmd, args=None, fn=None, cwd=None):
    """Pretty-print a command line."""
    if fn is None:
        fn = print
    if args is None:
        args = []
    if cwd is None:
        cwd = os.getcwd()
    fn(f"  {cmd}")
    i = 0
    lines = []
    for item in args:
        # convert all args to str
        if not isinstance(item, str):
            if isinstance(item, bytes):
                item = item.decode("utf-8")
            item = str(item)
        ispath = False
        if item.find(os.path.sep) > -1:
            if os.path.dirname(item) == cwd:
                item = os.path.basename(item)
            ispath = True
        if sys.platform == "win32":
            item = sp.list2cmdline([item])
            if not item.startswith('"'):
                item = quote_args([item])[0]
        else:
            item = pipes.quote(item)
        lines.append(item)
        i += 1
    for line in lines:
        fn(
            textwrap.fill(
                line,
                80,
                expand_tabs=False,
                replace_whitespace=False,
                initial_indent="    ",
                subsequent_indent="      ",
            )
        )


class ThreadAbort(object):
    def __init__(self):
        self.event = mp.Event()

    def __bool__(self):
        return self.event.is_set()

    def __cmp__(self, other):
        if self.event.is_set() < other:
            return -1
        if self.event.is_set() > other:
            return 1
        return 0


class WorkerBase(object):
    def __init__(self):
        """Create and return a new base worker instance."""
        self.sessionlogfile = None
        self.subprocess_abort = False
        self.tempdir = None
        self._thread_abort = ThreadAbort()

    def create_tempdir(self):
        """Create a temporary working directory and return its path."""
        if not self.tempdir or not os.path.isdir(self.tempdir):
            # we create the tempdir once each calibrating/profiling run
            # (deleted by 'wrapup' after each run)
            if verbose >= 2:
                if not self.tempdir:
                    msg = "there is none"
                else:
                    msg = "the previous (%s) no longer exists" % self.tempdir
                print(f"{appname}: Creating a new temporary directory because", msg)
            try:
                self.tempdir = tempfile.mkdtemp(prefix=f"{appname}-")
            except Exception as exception:
                self.tempdir = None
                return Error(
                    f"Error - couldn't create temporary directory: {safe_str(exception)}"
                )
        return self.tempdir

    def isalive(self, subprocess=None):
        """Check if subprocess is still alive"""
        if not subprocess:
            subprocess = getattr(self, "subprocess", None)
        return subprocess and (
            (hasattr(subprocess, "poll") and subprocess.poll() is None)
            or (hasattr(subprocess, "isalive") and subprocess.isalive())
        )

    def log(self, *args, **kwargs):
        """Log to global logfile and session logfile (if any)"""
        msg = " ".join(safe_basestring(arg) for arg in args)
        fn = kwargs.get("fn", print)
        fn(msg)
        if self.sessionlogfile:
            self.sessionlogfile.write(msg + "\n")

    @property
    def thread_abort(self):
        return self._thread_abort

    @thread_abort.setter
    def thread_abort(self, abort):
        if abort:
            self._thread_abort.event.set()
        else:
            self._thread_abort.event.clear()

    def xicclu(
        self,
        profile,
        idata,
        intent="r",
        direction="f",
        order="n",
        pcs=None,
        scale=1,
        cwd=None,
        startupinfo=None,
        raw=False,
        logfile=None,
        use_icclu=False,
        use_cam_clipping=False,
        get_clip=False,
        show_actual_if_clipped=False,
        input_encoding=None,
        output_encoding=None,
    ):
        """Call xicclu, feed input floats into stdin, return output floats.

        input data needs to be a list of 3-tuples (or lists) with floats,
        alternatively a list of strings.
        output data will be returned in same format, or as list of strings
        if 'raw' is true.

        """
        with Xicclu(
            profile,
            intent,
            direction,
            order,
            pcs,
            scale,
            cwd,
            startupinfo,
            use_icclu,
            use_cam_clipping,
            logfile,
            self,
            show_actual_if_clipped,
            input_encoding,
            output_encoding,
        ) as xicclu:
            xicclu(idata)
        return xicclu.get(raw, get_clip)


class Xicclu(WorkerBase):
    def __init__(
        self,
        profile,
        intent="r",
        direction="f",
        order="n",
        pcs=None,
        scale=1,
        cwd=None,
        startupinfo=None,
        use_icclu=False,
        use_cam_clipping=False,
        logfile=None,
        worker=None,
        show_actual_if_clipped=False,
        input_encoding=None,
        output_encoding=None,
        convert_video_rgb_to_clut65=False,
        verbose=1,
    ):
        if not profile:
            raise Error("Xicclu: Profile is %r" % profile)
        WorkerBase.__init__(self)
        self.scale = scale
        self.convert_video_rgb_to_clut65 = convert_video_rgb_to_clut65
        self.logfile = logfile
        self.worker = worker
        self.temp = False
        utilname = "icclu" if use_icclu else "xicclu"
        xicclu = get_argyll_util(utilname)
        if not xicclu:
            raise Error(lang.getstr("argyll.util.not_found", utilname))
        if not isinstance(profile, (CGATS.CGATS, ICCP.ICCProfile)):
            if profile.lower().endswith(".cal"):
                profile = CGATS.CGATS(profile)
            else:
                profile = ICCP.ICCProfile(profile)
        is_profile = isinstance(profile, ICCP.ICCProfile)
        if (
            is_profile
            and profile.version >= 4
            and not profile.convert_iccv4_tags_to_iccv2()
        ):
            raise Error(
                "\n".join(
                    [lang.getstr("profile.iccv4.unsupported"), profile.getDescription()]
                )
            )
        if not profile.fileName or not os.path.isfile(profile.fileName):
            if profile.fileName:
                prefix = os.path.basename(profile.fileName)
            elif is_profile:
                prefix = (
                    make_filename_safe(profile.getDescription(), concat=False)
                    + profile_ext
                )
            else:
                # CGATS (.cal)
                prefix = "cal"
            prefix += "-"
            if not cwd:
                cwd = self.create_tempdir()
                if isinstance(cwd, Exception):
                    raise cwd
            fd, profile.fileName = tempfile.mkstemp("", prefix, dir=cwd)
            with os.fdopen(fd, "wb") as stream:
                profile.write(stream)
            self.temp = True
        elif not cwd:
            cwd = os.path.dirname(profile.fileName)
        profile_basename = os.path.basename(profile.fileName)
        profile_path = profile.fileName
        if sys.platform == "win32":
            profile_path = win32api.GetShortPathName(profile_path)
        self.profile_path = safe_str(profile_path)
        if sys.platform == "win32" and not startupinfo:
            startupinfo = sp.STARTUPINFO()
            startupinfo.dwFlags |= sp.STARTF_USESHOWWINDOW
            startupinfo.wShowWindow = sp.SW_HIDE
        xicclu = safe_str(xicclu)
        cwd = safe_str(cwd)
        self.verbose = verbose
        args = [xicclu, "-v%i" % verbose, "-s%s" % scale]
        self.show_actual_if_clipped = False
        if utilname == "xicclu":
            if (
                is_profile
                and show_actual_if_clipped
                and "A2B0" in profile.tags
                and ("B2A0" in profile.tags or direction == "if")
            ):
                args.append("-a")
                self.show_actual_if_clipped = True
            if use_cam_clipping:
                args.append("-b")
            if get_argyll_version("xicclu") >= [1, 6]:
                # Add encoding parameters
                # Note: Not adding -e -E can cause problems due to unitialized
                # in_tvenc and out_tvenc variables in xicclu.c for Argyll 1.6.x
                if not input_encoding:
                    input_encoding = "n"
                if not output_encoding:
                    output_encoding = "n"
                args += [
                    "-e" + safe_str(input_encoding),
                    "-E" + safe_str(output_encoding),
                ]
        args.append("-f" + direction)
        self.output_scale = 1.0
        if is_profile:
            if profile.profileClass not in (b"abst", b"link"):
                if intent:
                    args.append(f"-i{intent}")
                if order != "n":
                    args.append("-o" + order)
            if profile.profileClass != b"link":
                if direction in ("f", "ib") and (
                    pcs == "x" or (profile.connectionColorSpace == b"XYZ" and not pcs)
                ):
                    # In case of forward lookup with XYZ PCS, use 0..100 scaling
                    # internally so we get extra precision from xicclu for the
                    # decimal part. Scale back to 0..1 later.
                    pcs = "X"
                    self.output_scale = 100.0
                if pcs:
                    args.append("-p" + pcs)
        args.append(self.profile_path)
        if debug or verbose > 1:
            self.sessionlogfile = LogFile(
                profile_basename + ".xicclu", os.path.dirname(profile.fileName)
            )
            if is_profile:
                profile_act = ICCP.ICCProfile(profile.fileName)
                self.sessionlogfile.write(
                    "Profile ID %s (actual %s)"
                    % (hexlify(profile.ID), hexlify(profile_act.calculateID(False)))
                )
            if cwd:
                self.log(lang.getstr("working_dir"))
                indent = "  "
                for name in cwd.split(os.path.sep):
                    self.log(
                        textwrap.fill(
                            name + os.path.sep,
                            80,
                            expand_tabs=False,
                            replace_whitespace=False,
                            initial_indent=indent,
                            subsequent_indent=indent,
                        )
                    )
                    indent += " "
                self.log("")
            self.log(lang.getstr("commandline"))
            printcmdline(xicclu, args[1:], fn=self.log, cwd=cwd)
            self.log("")
        self.startupinfo = startupinfo
        self.args = args
        self.cwd = cwd
        self.spawn()

    def spawn(self):
        self.closed = False
        self.output = []
        self.errors = []
        self.stdout = tempfile.SpooledTemporaryFile()
        self.stderr = tempfile.SpooledTemporaryFile()
        self.subprocess = sp.Popen(
            self.args,
            stdin=sp.PIPE,
            stdout=self.stdout,
            stderr=self.stderr,
            cwd=self.cwd,
            startupinfo=self.startupinfo,
        )

    def devi_devip(self, n):
        if n > 236 / 256.0:
            n = colormath.convert_range(n, 236 / 256.0, 1, 236 / 256.0, 255 / 256.0)
        return VidRGB_to_cLUT65(eeColor_to_VidRGB(n))

    def __call__(self, idata):
        if not isinstance(idata, str):
            verbose = self.verbose
            if self.convert_video_rgb_to_clut65:
                devi_devip = self.devi_devip
            else:
                devi_devip = lambda v: v
            scale = float(self.scale)
            idata = list(idata)  # Make a copy
            for i, v in enumerate(idata):
                if isinstance(v, (float, int)):
                    self([idata])
                    return
                if not isinstance(v, str):
                    if verbose:
                        for n in v:
                            if not isinstance(n, (float, int)):
                                raise TypeError(
                                    "xicclu: Expecting list of "
                                    "strings or n-tuples with "
                                    "floats"
                                )
                    idata[i] = " ".join(str(devi_devip(n / scale) * scale) for n in v)
        else:
            idata = idata.splitlines()
        numrows = len(idata)
        chunklen = 1000
        i = 0
        p = self.subprocess
        prevperc = -1
        while True:
            # Process in chunks to prevent broken pipe if input data is too
            # large
            if getattr(sys, "_sigbreak", False) and not self.subprocess_abort:
                self.subprocess_abort = True
                print("Got SIGBREAK, aborting subprocess...")
            if self.subprocess_abort or self.thread_abort:
                if p.poll() is None:
                    p.stdin.write(b"\n")
                    p.stdin.close()
                    p.wait()
                raise Info(lang.getstr("aborted"))
            if p.poll() is None:
                # We don't use communicate() because it will end the
                # process
                joined_data = "\n".join(idata[chunklen * i: chunklen * (i + 1)]) + "\n"
                p.stdin.write(joined_data.encode())
                p.stdin.flush()
            else:
                # Error
                break
            perc = round(chunklen * (i + 1) / float(numrows) * 100)
            if perc > prevperc and self.logfile:
                self.logfile.write("\r%i%%" % min(perc, 100))
                prevperc = perc
            if chunklen * (i + 1) > numrows - 1:
                break
            i += 1

    def __enter__(self):
        return self

    def __exit__(self, etype=None, value=None, tb=None):
        self.exit()
        if tb:
            return False

    def close(self, raise_exception=True):
        if self.closed:
            return
        p = self.subprocess
        if p.poll() is None:
            try:
                p.stdin.write(b"\n")
            except IOError:
                pass
            p.stdin.close()
        p.wait()
        self.stdout.seek(0)
        self.output = self.stdout.readlines()
        self.stdout.close()
        self.stderr.seek(0)
        self.errors = self.stderr.readlines()
        self.stderr.close()
        if self.sessionlogfile and self.errors:
            self.sessionlogfile.write("\n".join(self.errors))
        if self.logfile:
            self.logfile.write("\n")
        self.closed = True
        if p.returncode and raise_exception:
            # Error
            raise IOError("\n".join(self.errors))

    def exit(self, raise_exception=True):
        self.close(raise_exception)
        if self.temp and os.path.isfile(self.profile_path):
            os.remove(self.profile_path)
            if self.tempdir and not os.listdir(self.tempdir):
                try:
                    shutil.rmtree(self.tempdir, True)
                except Exception as exception:
                    print(
                        "Warning - temporary directory '%s' could not be removed: %s"
                        % (self.tempdir, exception)
                    )

    def get(self, raw=False, get_clip=False, output_format=None, reverse=False):
        if raw:
            if self.sessionlogfile:
                self.sessionlogfile.write("\n".join(self.output))
                self.sessionlogfile.close()
            return self.output
        parsed = []
        j = 0
        verbose = self.verbose
        scale = float(self.scale)
        output_scale = float(self.output_scale)
        if self.convert_video_rgb_to_clut65:
            devop_devo = VidRGB_to_eeColor
        else:
            devop_devo = lambda v: v

        fmt = ""
        maxv = ""
        if output_format:
            fmt = output_format[0]
            maxv = output_format[1]
        # Interesting: In CPython, testing for 'if not x' is slightly quicker
        # than testing for 'if x'. (EOY: Yeah I measured it is ~3% faster)
        # Also, struct.pack is faster if the second argument is passed as an integer.
        clip = None
        for i, line in enumerate(self.output):
            if verbose:
                line = line.strip()
                if line.startswith(b"["):
                    if parsed and get_clip and self.show_actual_if_clipped:
                        parts = line.strip(b"[]").split(b",")
                        actual = [float(v) for v in parts[0].split()[1:4]]  # Actual CIE
                        actual.append(float(parts[1].split()[-1]))  # deltaE
                        parsed[-1].append(actual)
                    elif self.sessionlogfile:
                        self.sessionlogfile.write(line)
                    continue
                elif b"->" not in line:
                    if self.sessionlogfile and line:
                        self.sessionlogfile.write(line)
                    continue
                elif self.sessionlogfile:
                    self.sessionlogfile.write("#%i %s" % (j, line))
                parts = line.split(b"->")[-1].strip().split()
                clip = parts.pop() == b"(clip)"
                if clip:
                    parts.pop()
                j += 1
            else:
                parts = line.split()
            if reverse:
                parts = reversed(parts)
            if not output_format:
                out = [devop_devo(float(v) / output_scale) for v in parts]
                if get_clip and not self.show_actual_if_clipped:
                    out.append(clip)
            else:
                out = b"".join(
                    struct.pack(fmt, int(round(devop_devo(float(v) / scale) * maxv)))
                    for v in parts
                )
            parsed.append(out)
        if self.sessionlogfile:
            self.sessionlogfile.close()
        return parsed

    @property
    def subprocess_abort(self):
        if self.worker:
            return self.worker.subprocess_abort
        return False

    @subprocess_abort.setter
    def subprocess_abort(self, v):
        pass

    @property
    def thread_abort(self):
        if self.worker:
            return self.worker.thread_abort
        return None

    @thread_abort.setter
    def thread_abort(self, v):
        pass


class MP_Xicclu(Xicclu):
    def __init__(
        self,
        profile,
        intent="r",
        direction="f",
        order="n",
        pcs=None,
        scale=1,
        cwd=None,
        startupinfo=None,
        use_icclu=False,
        use_cam_clipping=False,
        logfile=None,
        worker=None,
        show_actual_if_clipped=False,
        input_encoding=None,
        output_encoding=None,
        output_format=None,
        reverse=False,
        output_stream=None,
        convert_video_rgb_to_clut65=False,
        verbose=1,
    ):
        WorkerBase.__init__(self)
        self.logfile = logfile
        self.worker = worker
        self.output_stream = output_stream
        self._in = []
        self._args = (
            profile.fileName,
            intent,
            direction,
            order,
            pcs,
            scale,
            cwd,
            startupinfo,
            use_icclu,
            use_cam_clipping,
            None,
            show_actual_if_clipped,
            input_encoding,
            output_encoding,
            lang.getstr("aborted"),
            output_format,
            reverse,
            convert_video_rgb_to_clut65,
            verbose,
        )
        self._out = []
        num_cpus = mp.cpu_count()
        if isinstance(profile.tags.get("A2B0"), ICCP.LUT16Type):
            size = profile.tags.A2B0.clut_grid_steps
            self.num_workers = min(max(num_cpus, 1), size)
            if num_cpus > 2:
                self.num_workers = int(self.num_workers * 0.75)
            self.num_batches = size // 9
        else:
            if num_cpus > 2:
                self.num_workers = 2
            else:
                self.num_workers = num_cpus
            self.num_batches = 1

    def __call__(self, idata):
        self._in.append(idata)

    def close(self, raise_exception=True):
        pass

    def exit(self, raise_exception=True):
        pass

    def spawn(self):
        pass

    def get(self, raw=False, get_clip=False, output_format=None, reverse=False):
        for slices in pool_slice(
            _mp_xicclu,
            self._in,
            self._args,
            {},
            self.num_workers,
            self.thread_abort,
            self.logfile,
            num_batches=self.num_batches,
        ):
            if self.output_stream:
                for row in slices:
                    self.output_stream.write(row)
            else:
                self._out.extend(slices)
        return self._out
